# -*- test-case-name: safe.test.boolean -*-

"""
The classes in this module provide facilities to build boolean expressions.
These expressions can then be used as filters to prune either the parameter
space, or query results for different design points.

For example: ::

    >>> t1 = Term("param1", 1)
    >>> t2 = Term("param2", 2)
    >>> t3 = Term("param3", 3)
    >>> t4 = Term("param4", 4)
    >>> exp = (t1 & t2) | (t3 ^ -t4)
    >>> exp.evaluate({"param1": 1, "param2": 2, "param3": 3, "param4": 4})
    True
    >>> exp.evaluate({"param1": 123, "param2": 2, "param3": 456, "param4": 4})
    False
"""

__author__ = "Bryan C. Ward"
__email__ = "bryan.ward@bucknell.edu"

class BooleanTerm(object):
    """ An abstract class which provides the syntax we want to use.

    Subclasses of :py:class:`BooleanTerm` can use ``-term`` to indicate unary
    negation. Then they can use the binary operations ``&``, ``|`` and ``^``.
    The python keyword ``not`` cannot be overridden, so the unary negation
    symbol is used in its place.

    The result of any binary operation is an :py:class:`Expression`.  In this
    manner, a tree is built for the boolean expression.

    """

    def __neg__(self):
        """ The logical negation of the :py:class:`BooleanTerm`."""
        self.invert = not self.invert
        return self

    def __inv__(self):
        """ The logical negation of the :py:class:`BooleanTerm`."""
        self.invert = not self.invert
        return self

    def __and__(self, other):
        """ The binary operator `and` of two :py:class:`BooleanTerm`s."""
        return Expression(self, "and", other)

    def __or__(self, other):
        """ The binary operator `or` of two :py:class:`BooleanTerm`s."""
        return Expression(self, "or", other)

    def __xor__(self, other):
        """ The binary operator `xor` of two :py:class:`BooleanTerm`s."""
        return Expression(self, "xor", other)

class Expression(BooleanTerm):
    """ This class represents a binary expression.

    An expression is built from two operands, :py:attr:`op1` and :py:attr:`op2`.
    An operand can be another :py:class:`Expression` or a :py:class:`Term`. When
    an :py:class:`Expression` is evaluated on an input, the tree formed from
    these Expressions and Terms is traversed to evaluate if the input is true or
    false given the expression.
    
    """

    def __init__(self, op1, op, op2, invert = False):
        self.op1 = op1
        self.op = op
        self.op2 = op2
        self.invert = invert

    def evaluate(self, input, default = False):
        """ Evaluate whether the input is true or false.

        :param input: a dictionary with the values to evaluate the expression
            on.
        :param default: the boolean value returned if an input key is not found.
            The default is False
        :rtype: True or False

        """
        if self.op == "and":
            retval = self.op1.evaluate(input, default) and self.op2.evaluate(input, default)
        elif self.op == "or":
            retval = self.op1.evaluate(input, default) or self.op2.evaluate(input, default)
        elif self.op == "xor":
            retval = self.op1.evaluate(input, default) ^ self.op2.evaluate(input, default)
        else:
            raise BooleanOperatorError(self.op)

        return retval ^ self.invert

    def __str__(self):
        retval = '%s %s %s' % (self.op1.__str__(), self.op.__str__(), self.op2.__str__())
        return retval
        

class Term(BooleanTerm):

    def __init__(self, key, value, invert = False):
        self.key = key
        self.value = value
        self.invert = invert

    def evaluate(self, input, default = False):
        """ Evaluate whether the input is true or false.

        :param input: a dictionary with the values to evaluate the expression
            on.
        :param default: the boolean value returned if an input key is not found.
            The default is False
        :rtype: True or False

        """

        if input.has_key(self.key):
            retval = ( input[self.key] == self.value )
        else:
            retval = default
        
        return retval ^ self.invert

class AndList(BooleanTerm):
    """ An :py:class:`AndList` allows the user to specify several and terms with
    one class.

    """

    def __init__(self, constraints, invert = False):
        """ Constructor for the :py:class:`OrList`.

        :param constraints: A dictionary mapping a key to a value. If any one of
        the conditions hold, the :py:class:`OrList` evaluates to True, otherwise
        False.
        :rtype: True or False

        """
        self.list = []
        self.invert = invert
        for key, value in constraints.iteritems():
            self.list.append(Term(key,value))

    def evaluate(self, input, default = False):
        """ Evaluate whether the input is true or false.

        :param input: a dictionary with the values to evaluate the expression
            on.
        :param default: the boolean value returned if an input key is not found.
            The default is False
        :rtype: True or False

        """

        retval = True
        for term in self.list:
            if not term.evaluate(input, default):
                retval = False
                break
        return retval ^ self.invert

class OrList(BooleanTerm):
    """ An :py:class:`OrList` allows the user to specify several or terms with
    one class.

    """

    def __init__(self, constraints, invert = False):
        """ Constructor for the :py:class:`OrList`.

        :param constraints: A dictionary mapping a key to a value. If any one of
        the conditions hold, the :py:class:`OrList` evaluates to True, otherwise
        False.
        :rtype: True or False

        """

        self.list = []
        self.invert = invert
        for key, value in constraints.iteritems():
            self.list.append(Term(key,value))

    def evaluate(self, input, default = False):
        """ Evaluate whether the input is true or false.

        :param input: a dictionary with the values to evaluate the expression
            on.
        :param default: the boolean value returned if an input key is not found.
            The default is False
        :rtype: True or False

        """

        retval = False
        for term in self.list:
            if term.evaluate(input, default):
                retval = True
                break
        return retval ^ self.invert
