/*
 * samplestat.cc :- collecting statistics
 */

/*
 * Copyright (c) 1998-2002 Dartmouth College
 *
 * Permission is hereby granted, free of charge, to any individual or 
 * institution obtaining a copy of this software and associated 
 * documentation files (the "Software"), to use, copy, modify, and 
 * distribute without restriction, provided that this copyright and 
 * permission notice is maintained, intact, in all copies and supporting 
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DARTMOUTH COLLEGE BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <cmath>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iostream>

#include "samplestat.h"

SampleStat::SampleStat() 
{
  _nentries = 0.0;
  _min = _max = 0.0;
  _mean = _sumsqr = 0.0;
  set_conf_level();
}

SampleStat::~SampleStat() {}

inline double SampleStat::get_sample(int index) {
  return samples[index];
}

void SampleStat::record(double value) 
{
  _nentries += 1.0;
  if(_nentries == 1.0) _min = _max = value;
  else {
    if(value < _min) _min = value;
    if(value > _max) _max = value;
  }
  _sumsqr += (value-_mean)*(value-_mean)*(_nentries-1)/_nentries;
  _mean += (value-_mean)/_nentries;

  samples.push_back(value);
}

double SampleStat::window_average(int winpos, int winsize) {

  double winsum(0);

  for (int i=0; i < winsize; i++)
    winsum = winsum + samples[winpos+i];

  return winsum/winsize;
}

vector<double> *SampleStat::moving_averages(int winsize) {

  vector<double> *averages;

  averages = new vector<double>;

  for (int i=0; i < ((int) samples.size()) - winsize; i++) {
    cout << "win average[" << i << "] = " << window_average(i,winsize) << endl;
    cout.flush();

    (*averages)[i] = window_average(i, winsize);

    cout << "passei!" << endl; cout.flush();

  }

  return averages;
}

void SampleStat::report(FILE *outf) 
{
  fprintf(outf, "==================\n");
  fprintf(outf, "sample set:\n");
  fprintf(outf, "==================\n");
  fprintf(outf, "\tsize=%ld [min=%f, max=%f]\n",
	  long(_nentries), _min, _max);
  fprintf(outf, "\tmean=%f\n", _mean);
  fprintf(outf, "\tvariance=%f (stdev=%f)\n", 
	  variance(), stdev());
  double x = conf_intv_half_width();
  fprintf(outf, "\t%g%% confidence interval (assuming iid):\n", _conf_level*100);
  fprintf(outf, "\t\t%f +/- %f [%f, %f]\n", _mean, x, _mean-x, _mean+x);
  fprintf(outf, "\n");
}

double SampleStat::min() { return _min; }

double SampleStat::max() { return _max; }

double SampleStat::range() { return _max-_min; }

double SampleStat::size() { return _nentries; }

double SampleStat::mean() { return _mean; }

double SampleStat::variance() { return _sumsqr/_nentries; }

double SampleStat::stdev() { return sqrt(variance()); }

double SampleStat::var_coef() 
{ 
  if(_mean == 0.0) return 0.0;
  else return stdev()/_mean;
}

void SampleStat::set_conf_level(double conf)
{
  _conf_level = conf;
}

double SampleStat::conf_level() { return _conf_level; }

double SampleStat::conf_intv_half_width() 
{
  if(_nentries <= 1.0) return 0.0;
  double u = 1.0-0.5*(1.0-_conf_level);
  double t = student_idf(long(_nentries)-1, u);
  return t*stdev()/sqrt(_nentries-1);
}

/*----------------------------------------------------------------------------*/
/*   This function computes the pth upper quantile of the standard normal     */
/*   distribution  (i.e., the value of z for which the area under the curve   */
/*   from z to +infinity is equal to p).                                      */
/*                                                                            */
/*   This algorithm has been taken from:                                      */
/*                                                                            */
/*   M. H. MacDougall, Simulating Computer Systems, MIT Press, 1987, p. 276.  */
/*   G. S. Fishman, Principles of Discrete Event Simulation, Wiley, 1978.     */
/*   C. Hastings, Jr., Approximations for Digital Computers, Princeton, 1955. */
/*----------------------------------------------------------------------------*/

static double normal_quantile(double p)
{
  double q, z1, n, d;
   
  if (p > 0.5)  q = 1.0 - p;
  else q = p;
  z1 = sqrt (-2.0 * log(q));
  n = (0.010328 * z1 + 0.802853) * z1 + 2.515517;
  d = ((0.001308 * z1 + 0.189269) * z1 + 1.43278) * z1 + 1.0;
  z1 = z1 - n / d;
  if (p > 0.5)  z1 = -z1;
  return (z1);
}  /* normal_quantile */

/*----------------------------------------------------------------------------*/
/*   This function computes the pth upper quantile of the t distribution      */
/*   having ndf degrees of freedom  (i.e., the value of t for which the area  */
/*   under the curve from t to +infinity is equal to p).                      */
/*                                                                            */
/*   This algorithm has been taken from:                                      */
/*                                                                            */
/*   M. H. MacDougall, Simulating Computer Systems, MIT Press, 1987, p. 276.  */
/*   G. S. Fishman, Principles of Discrete Event Simulation, Wiley, 1978.     */
/*----------------------------------------------------------------------------*/

static double t_quantile(double p, long ndf)
{
  long i;
  double z1, z2, h[4], x;
   
  if (ndf == 0)  return (0.0);
  x = 0.0;
  z1 = fabs(normal_quantile(p));
  z2 = z1 * z1;
  h[0] = 0.25 * z1 * (z2 + 1.0);
  h[1] = 0.010416667 * z1 * ((5.0 * z2 + 16.0) * z2 + 3.0);
  h[2] = 0.002604167 * z1 * (((3.0 * z2 + 19.0) * z2 + 17.0) * z2 - 15.0);
  h[3] = z1 * ((((79.0 * z2 + 776.0) * z2 + 1482.0) * z2 -1920.0) * z2 -945.0);
  h[3] = h[3] * 0.000010851;
  for (i=3; i>=0; i--)
    x = (x + h[i]) / (double) ndf;
  z1 = z1 + x;
  if (p > 0.5)  z1 = -z1;
  return (z1);
}  /* t_quantile */

double SampleStat::student_idf(long n, double u)
{
  return t_quantile(1.0-u, n);
}

SampleHistogram::SampleHistogram(long nbuckets, double lo, double hi) :
  _nbuckets(nbuckets), _low(lo), _high(hi) 
{
  if (lo >= hi || nbuckets <= 0) {
    printf("invalid SampleHistogram: num_buckets(%ld), low(%f), high(%f)\n",
	       nbuckets, lo, hi);
    exit(-1);
  }

  _width = (hi-lo)/nbuckets;
  _frequency = new double[nbuckets];
  memset(_frequency, 0, nbuckets*sizeof(double));
  _nentries = 0.0;
  _underliers = _overliers = 0.0;
  _mature = 1;
  _mean_computed = _var_computed = 0;
}

SampleHistogram::SampleHistogram(long bufsiz) 
{
  if(bufsiz < 2) {
    printf("invalid SampleHistogram: bufsiz(%ld)\n", bufsiz);
    exit(-1);
  }

  _bufsiz = bufsiz;
  _buffer = new double[_bufsiz];
  long x = bufsiz; 
  _nbuckets = 0; while(x > 1) { x >>= 1; _nbuckets++; }
  _frequency = new double[_nbuckets];
  memset(_frequency, 0, _nbuckets*sizeof(double));
  _nentries = 0.0;
  _underliers = _overliers = 0.0;
  _mature = 0;
  _mean_computed = _var_computed = 0;
}

SampleHistogram::~SampleHistogram() 
{ 
  if(!_mature) delete[] _buffer;
  delete[] _frequency; 
}

void SampleHistogram::record(double value) 
{
  _nentries += 1.0;
  if(!_mature) {
    if(_nentries == 1.0) _low = _high = value;
    if(_nentries <= _bufsiz) {
      _buffer[long(_nentries)-1] = value;
      if(_high < value) _high = value;
      else if(value < _low) _low = value;
    } 
    if(_nentries == _bufsiz) grow_up();
  } else {
    _mean_computed = _var_computed = 0;
    if(_low <= value && value <= _high) {
      long i = long((value-_low)/_width);
      if(i == _nbuckets) i--;
      _frequency[i]++;
    }
    else if(value < _low) _underliers++;
    else _overliers++;
  }
}

#define MAXBAR 20

void SampleHistogram::report(FILE *outf) 
{
  fprintf(outf, "==================\n");
  fprintf(outf, "SampleHistogram:\n");
  fprintf(outf, "==================\n");
  fprintf(outf, "\tnum_buckets=%ld [low=%f, high=%f) => width(%f)\n",
	  _nbuckets, _low, _high, _width);
  fprintf(outf, "\tnum_observations=%ld\n", long(_nentries));
  fprintf(outf, "\tnum_outliers=%ld\n", long(num_outliers()));
  if(_nentries == 0.0) return;

  fprintf(outf, "\tmean=%f\n", mean());
  double x = variance();
  fprintf(outf, "\tvariance=%f (stdev=%f)\n", x, sqrt(x));

  fprintf (outf, "                                                cumulative\n");
  fprintf (outf, "      lower limit     frequency    proportion   proportion\n");
  fprintf (outf, "\n");

  register long i, j;

  /* find fullest bucket */
   
  double maxp = 0.0;
  double p;
  for(i=0; i<_nbuckets; i++) {
    p = _frequency[i]/_nentries;
    if(p > maxp) maxp = p;
  }
  p = _underliers/_nentries;
  if(p > maxp) maxp = p;
  p = _overliers/_nentries;
  if(p > maxp) maxp = p;

  /* truncate empty buckets */
   
  long start_bucket = 0;
  long stop_bucket = _nbuckets-1;
  if(_underliers == 0)
    while(start_bucket < _nbuckets &&
	  _frequency[start_bucket] == 0.0) start_bucket++;
  if(_overliers == 0)
    while(stop_bucket >= 0 && 
	  _frequency[stop_bucket] == 0.0) stop_bucket--;

   /* loop over buckets */
      
  double cumulative = 0.0;
  if(_underliers > 0) {
    p = _underliers/_nentries;
    cumulative += p;
    fprintf (outf, "     <%11.5f   %11ld     %8.6f     %8.6f   ",
	     _low, long(_underliers), p, cumulative);
    long bar = long(p/maxp*MAXBAR+0.5);
    if(bar == 0 && p > 0.0) fprintf (outf, ".");
    else for(j=1; j<=bar; j++) fprintf (outf, "*");
    fprintf (outf, "\n");
  }
  for(i=start_bucket; i<=stop_bucket; i++) {
    p = _frequency[i]/_nentries;
    cumulative += p;
    fprintf (outf, "      %11.5f   %11ld     %8.6f     %8.6f   ",
	     _low+i*_width, (long)_frequency[i], p, cumulative);
    long bar = long(p/maxp*MAXBAR+0.5);
    if(bar == 0 && p > 0.0) fprintf (outf, ".");
    else for(j=1; j<=bar; j++) fprintf (outf, "*");
    fprintf (outf, "\n");
  }
  if(_overliers > 0) {
    p = _overliers/_nentries;
    cumulative += p;
    fprintf (outf, "    >=%11.5f   %11ld     %8.6f     %8.6f   ",
	     _high, long(_overliers), p, cumulative);
    long bar = long(p/maxp*MAXBAR+0.5);
    if(bar == 0 && p > 0.0) fprintf (outf, ".");
    else for(j=1; j<=bar; j++) fprintf (outf, "*");
    fprintf (outf, "\n");
  }
  fprintf(outf, "\n");
}

#undef MAXBAR

long SampleHistogram::num_buckets() { return _nbuckets; }
  
double SampleHistogram::low() { return _low+0.5*_width; }

double SampleHistogram::high() { return _high-0.5*_width; }

double SampleHistogram::width() { return _width; }

double SampleHistogram::num_observations() { return _nentries; }

double SampleHistogram::num_outliers() 
{ 
  if(!_mature) grow_up();
  return _underliers+_overliers;
}
  
double SampleHistogram::frequency(long i) 
{ 
  if(i < 0 || i >= _nbuckets) return 0;
  else return _frequency[i]; 
}

double SampleHistogram::mean() 
{
  if(_mean_computed) return _mean;
  if(!_mature) grow_up();
  double m = 0;
  double x = _low+_width/2.0;
  for(long i=0; i<_nbuckets; i++) {
    m += _frequency[i]*x;
    x += _width;
  }
  m /= _nentries-num_outliers();
  _mean_computed = 1; _mean = m;
  return m;
}

double SampleHistogram::variance() 
{
  if(_var_computed) return _var;
  double m = mean();
  double v = 0;
  double x = _low+_width/2.0;
  for(long i=0; i<_nbuckets; i++) {
    v += _frequency[i]*(x-m)*(x-m);
    x += _width;
  }
  v /= _nentries-num_outliers();
  _var_computed = 1; _var = m;
  return v;
}

double SampleHistogram::stdev() { return sqrt(variance()); }

void SampleHistogram::grow_up() 
{
  _mature = 1;
  _width = (_high-_low)/_nbuckets;
  if(_high == _low) {
    printf("SampleHistogram premature (identical pre-samples)");
    _width = 1.0;
  }
  for(long i=0; i<_nentries; i++) {
    if(_low <= _buffer[i] && _buffer[i] <= _high) {
      long j = long((_buffer[i]-_low)/_width);
      if(j == _nbuckets) j--;
      _frequency[j]++;
    }
    else if(_buffer[i] < _low) _underliers++;
    else _overliers++;
  }
  delete[] _buffer;
}
