from safe.messages import *
from struct import *
import sys
from twisted.trial import unittest

class TestDecode(unittest.TestCase):

    goodCases = [
            (pack(Register._format,REGISTER,"uname"),Register),
            (pack(NextRequest._format,NEXT_REQUEST),NextRequest),
            (pack(NextReply._format % (10),
                NEXT_REPLY, 10, "<testing/>"),NextReply),
            (pack(Sample._format,SAMPLE,5.34,"node1","e2e",1.23),Sample),
            (pack(Finished._format,FINISHED),Finished),
            (pack(Terminate._format,TERMINATE),Terminate)
            ]

    badCases = [
            (pack(Register._format,REGISTER+1,"uname"),Register),
            (pack(NextRequest._format,NEXT_REQUEST+1),NextRequest),
            (pack(NextReply._format % (10),
                NEXT_REPLY+1, 10, "<testing/>"),NextReply),
            (pack(Sample._format,SAMPLE+1,5.34,"node1","e2e",1.23),Sample),
            (pack(Finished._format,FINISHED+1),Finished),
            (pack(Terminate._format,TERMINATE+1),Terminate)
            ]

    def testTypes(self):
        for data,type in self.goodCases:
            msg,rest = decode(data)
            self.assertEqual(msg,type.decode(data))

    def testBadOpcode(self):
        for data,type in self.badCases:
            self.assertRaises(MessageDecodeError,type.decode,data)

    def testEmpty(self):
        msg, data = decode("")
        self.assertEqual(msg,None)
        self.assertEqual(data,"")

    def testDecodeBadOpcode(self):
        self.assertRaises(MessageDecodeError,decode,"bad data")
        
    def testNotEnoughData(self):
        data = pack("!II",NEXT_REPLY,100)+"not enough xml"
        msg, ret_data = decode(data)
        self.assertEqual(msg,None)
        self.assertEqual(data,ret_data)

        data = pack("!I",REGISTER)+"not enough data"
        msg, ret_data = decode(data)
        self.assertEqual(msg,None)
        self.assertEqual(data,ret_data)

class TestRegister(unittest.TestCase):

    def setUp(self):
        self.reg = Register("uname")

    def testNew(self):
        self.assertEqual(self.reg.info,"uname")

    def testEncode(self):
        result = pack(Register._format,REGISTER,"uname")
        self.assertEqual(self.reg.encode(),result)

    def testDecode(self):
        testing = pack(Register._format,REGISTER,"uname")
        result = Register.decode(testing)
        self.assertEqual(result.info.strip('\x00'),"uname")

        testing = pack(Register._format,REGISTER+1,"uname")
        self.assertRaises(MessageDecodeError,Register.decode,testing)

class TestNextReply(unittest.TestCase):
    def testNew(self):
        test = NextReply("<testing/>")
        self.assertEqual(test.xml,"<testing/>")

    def testDecode(self):
        test = pack(NextReply._format % (10), NEXT_REPLY, 10, "<testing/>")
        result = NextReply.decode(test)
        self.assertEqual(result,NextReply("<testing/>"))

        test = pack(NextReply._format % (10), 182, 10, "<testing/>")
        self.assertRaises(MessageDecodeError,NextReply.decode,test)

class TestNextRequest(unittest.TestCase):

    def setUp(self):
        self.uut = NextRequest()

    def testEncode(self):
        self.assertEqual(self.uut.encode(),
                pack(NextRequest._format,NEXT_REQUEST))

class TestNextReply(unittest.TestCase):

    def setUp(self):
        self.uut = NextReply("<testing/>")

    def testEncode(self):
        enc = self.uut.encode()
        goal = pack(NextReply._format % (10),NEXT_REPLY,10,"<testing/>")
        self.assertEqual(enc,goal)

class TestSample(unittest.TestCase):

    def setUp(self):
        self.uut = Sample(1.23,"node1","e2e",4.56)

    def testNew(self):
        self.assertEqual(self.uut.timestamp,1.23)
        self.assertEqual(self.uut.object.strip('\x00'),"node1")
        self.assertEqual(self.uut.metric.strip('\x00'),"e2e")
        self.assertEqual(self.uut.value,4.56)

    def testDecode(self):
        test = pack(Sample._format,SAMPLE,5.34,"node1","e2e",1.23)
        result = Sample.decode(test)
        self.assertEqual(result.timestamp,5.34)
        self.assertEqual(result.object.strip('\x00'),"node1")
        self.assertEqual(result.metric.strip('\x00'),"e2e")
        self.assertEqual(result.value,1.23)

        test = pack(Sample._format, 182, 1.23,"node1","e2e",1.23)
        self.assertRaises(MessageDecodeError,Sample.decode,test)

    def testEncode(self):
        goal = pack(Sample._format,SAMPLE,1.23,"node1","e2e",4.56)
        self.assertEqual(self.uut.encode(),goal)

class TestFinished(unittest.TestCase):

    def setUp(self):
        self.uut = Finished()

    def testEncode(self):
        goal = pack(Finished._format,FINISHED)
        self.assertEqual(self.uut.encode(),goal)

class TestMessageDecodeError(unittest.TestCase):
    def testString(self):
        ex = MessageDecodeError("testing")
        self.assertEqual(str(ex),"testing")
