#!/usr/bin/env python

""" manager.py: The Experiment Execution Manager (EEM) classes. """

from twisted.internet.protocol import Protocol, Factory, ClientFactory
from twisted.protocols.basic import Int32StringReceiver
from twisted.internet import reactor, defer, threads
from twisted.application import internet
from twisted.internet.defer import Deferred, maybeDeferred, DeferredList
from datetime import  datetime
from fabric.colors import blue, yellow
import argparse
import time
import sys
import os
import json
import base64
import struct
import socket

#Django requires the safe directory to be in the python path
sys.path.append(os.path.join(os.path.abspath(os.path.dirname(__file__)),"web"))

#Import fabric that is used to ssh workers
from fabric.api import run,roles,execute
import fabfile

#Uses same settings as django
from global_analyzer import AnalysisManager, AnalysisFactory
from django.conf import settings
from django.db.models import Count, Avg
os.environ.setdefault("DJANGO_SETTINGS_MODULE","web.config.settings")

#Import database models from django
from web.safe.models import *

class AnalysisClientFactory(ClientFactory):
  """Class for communicating with the AnalysisManager used for global analysis
  of results. Overwritten so that the protocol is publicly accessible, allowing
  the sendString method to be called from the ExperimentManager when it needs
  to send an estimate to the AanalysisManager"""
  protocol = Int32StringReceiver
  def buildProtocol(self, address):
    proto = ClientFactory.buildProtocol(self, address)
    self.connectedProtocol = proto
    return proto




class ExperimentManager(Int32StringReceiver):
  """Manager for a single execution client.

  The class extends the :class:`Int32StringReceiver` class and overrides the
  :func:`dataReceived` method. This allows it to accept messages from the
  remote client in an asynchronous fashion. A single instance of this class
  manages a single client connection, so any state variables specific to this
  instance are for a given client. The ``self.factory`` variable is the
  object which holds any information for the aggregate of all
  :class:`ExperimentManager` instances."""

  def __init__(self, factory):
    self.factory = factory

    #Data buffers
    self.double_result_buffer = []
    self.integer_result_buffer = []
    self.boolean_result_buffer = []
    metrics = self.factory.experiment.exp_description.metric_set.order_by('id')
    r = list(metrics[:1])
    if r:
      self.pos_offset = int(r[0].id)
    else:
      self.pos_offset = None
    #Database worker object representing this worker
    self.worker = None
    self.used_workers = 0
    self.max_workers = 0

    self.prev_count = {}
    #Database simulation run object representing the current simulation run
    #assigned to this worker
    self.simulation_run = []
    self.dp = 0
    self.client = AnalysisClientFactory()
    reactor.connectTCP(socket.gethostname(), 8124, self.client)

    #Result sets in order of receiving them. The order is important as the worker will only refer it by the index
    self.result_sets = {}
    self.done = 0
    #Switch demuxing all received messages
    self.handlers = {
    "register" : self._handleRegister,
    "dp_request" : self._handleDpRequest,
    "new_metric" : self._handleNewMetric,
    "result" : self._handleResult,
    "finished" : self._handleFinished,
    "dp_finished" : self._handleDpFinished}


  def stringReceived(self,raw_data):
    """Method invoked when data is received on the network."""
    #Decode json data to a dictionary
    data = json.loads(raw_data)
    #Call the correct handler function
    if data['type'] in self.handlers.keys():
      self.handlers[data['type']](data)
    else:
      raise Exception("Invalid message received: " + str(data))

  def connectionMade(self):
    """Method invoked when a connection to the client is made."""
    print("New connected client ")

  def connectionLost(self,reason):
    """Method invoked when the connection to the client is lost.

      When the connection is lost, ``self`` should be removed from the list
      of active protocols. In the case that all protocols have terminated
      their connections and there are no design points left to simulate, the
      execution manager can terminate.

      """
    #sys.stderr.write("Connection Lost: %s\n" % self.transport.getPeer())
    #sys.stderr.write(str(reason)+"\n")


  def _flushBuffers(self):
    """Constantly sending data to the database would add a lot of overhead,
    so we store the results in an array, and send them to the database when we've
    accumulated a certain amount or when a designpoint or simulationrun is 
    finished"""
    DoubleData.objects.bulk_create(self.double_result_buffer)
    IntegerData.objects.bulk_create(self.integer_result_buffer)
    BooleanData.objects.bulk_create(self.boolean_result_buffer)
    self.double_result_buffer = []
    self.integer_result_buffer = []
    self.boolean_result_buffer = []

  def _handleRegister(self,msg):
    """Method invoked when a register message is accepted from the
      network.

      The method will store the host information stored in the incoming
      message (msg) in the database.

      """

    print("Handling REGISTER")
    #The user must not be logged before
    if self.worker:
      sys.stderr.write("Worker already logged in, ignoring register message\n")
      return

    #Try to authenticate with the database
    try:
      self.worker = Worker.objects.get(auth_key__exact = msg['auth_key'])
      print("Worker successfully authenticated")
      self.max_workers = int(self.worker.children)
      self.simulation_run = [None for x in range(self.max_workers)]
      self.dp = list(self.factory.experiment.designpoint_set.all().order_by('pk')[:1])[0].id


    #If auth_key is not found in the database
    except Worker.DoesNotExist, Worker.MultipleObjectsReturned:
      sys.stderr.write("Invalid authentication, disconnecting\n")
      self.transport.loseConnection()


  def _handleDpRequest(self,msg):
    """Method invoked when a :class:`NextRequest` message is accepted.

      The method will get the next :class:`DesignPoint` to dispatch, using a
      database query, and then it will send that DesignPoint to the client."""
    print("Received DP Request: " + str(msg))
    #Check if the worker is logged in
    if not self.worker:
      sys.stderr.write("Not logged in, not authorized\n")
      self.transport.loseConnection()
      return

    proc_id = int(msg['id'])

    #Check if it doesnt already have one simulation going on, or it's already finished with one simulation
    if proc_id >= len(self.simulation_run) or (self.simulation_run[proc_id] and not self.simulation_run[proc_id].end_time ):
      print(str(self.simulation_run[proc_id].id) + '  ' + str(len(self.simulation_run)))
      sys.stderr.write("Worker already has simulation going on\n")
      sys.stderr.flush()
      sys.stdout.flush()
      return

    #May need to change if/when we remove replications in favor of setting number of replications to the number of workers/children
    query = DesignPoint.objects.filter(experiment_id__exact = self.factory.experiment.id)\
                               .annotate(total_runs = Count('simulationrun'))\
                               .filter(total_runs__lt = self.factory.experiment.replications, finished = 0)

    try:
      print(self.dp)
      #If there are no design_points left
      if self.dp <= -1 :
        print('No Valid Design Points Found. Finishing Experiment.')
        #to avoid repeat messages and buffer flushing, use self.done to
        #check if we've already sent the "finished"" message
        if not self.done:
          self._flushBuffers()
          self.sendString(json.dumps({"type":"finished"}))
          self.used_workers -= 1
          self.done = True

      else:
        #get the current design_point
        design_point = query.get(pk = self.dp)        
        print("Found design point (ID " + str(design_point.id) + ") with " + str(design_point.total_runs) + " runs so far")
        prev_dp = 0
        #If we previously had a design_point, keep track of it so we can kill any
        if self.simulation_run[proc_id]:
         prev_dp = self.simulation_run[proc_id].design_point.id
        #Create the new simulationRun object
        self.simulation_run[proc_id] = SimulationRun()
        self.simulation_run[proc_id].worker = self.worker
        self.simulation_run[proc_id].design_point = design_point
        self.simulation_run[proc_id].start_time = datetime.now()
        #Save it to the database
        self.simulation_run[proc_id].save()

        #Now get the factors and levels, report it to the simulation client (worker)
        #This dictionary will receive a pair of {factor: level}
        input = {}

        #It is necessary to get all factors AND levels separately because some factors
        #can have no level associate with it (only the default value).

        #Pass in the correct run number to advance the seed of the RNG
        input['RngRun'] = str(self.simulation_run[proc_id].id)

        #Getting all factors
        factors = Factor.objects.filter(exp_description = self.factory.experiment.exp_description)

        for factor in factors:
          input[factor.name] =  factor.getDefault()

        #Getting all levels
        levels = DesignPoint.objects.select_related("levels__factor__name", "levels__factor__id").get(pk = design_point.id).level_set.all()

        for level in levels:
          input[level.factor.name] = level.getValue()

        #Putting everything in a message and send it to the worker 0.
        self.used_workers += 1
        self.sendString(json.dumps({"type": "design_point", "input": input, 'id': msg['id'], 'dp_id':design_point.id, 'prev_dp_id':prev_dp}))

    except DesignPoint.DoesNotExist:
      #There is no design point for this experiment!!!
      sys.stderr.write("No design point available for this experiment\n")
      self.sendString(json.dumps({"type":"finished"}))

  def _handleFinished(self,msg):
    """Method called when a :class:`Finished` message is received."""
    print("Got finished, Used Workers =" + str(self.used_workers))
    
    if self.used_workers == 0:
      print("FINISHED received: " + str(msg))
      print("FINISHING CONNECTION")
      self.transport.loseConnection()
      self._flushBuffers()
      #If finished, make sure no one else sends finished method
      self.used_workers -= 1

    proc_id = int(msg['id'])
    self.simulation_run[proc_id].end_time = datetime.now()
    self.simulation_run[proc_id].isFinished = True
    self.simulation_run[proc_id].success = msg["success"]
    self.simulation_run[proc_id].save()


  def _handleDpFinished(self,msg):
    """Method called when a :class:`DpFinished` message is received.
    This represents when a worker has finished with a single simulationrun"""

    print("DPFINISHED received: " + str(msg))
    self.used_workers -= 1

    proc_id = int(msg['id'])
    self.simulation_run[proc_id].end_time = datetime.now()
    self.simulation_run[proc_id].isFinished = True
    self.simulation_run[proc_id].success = msg["success"]
    self.simulation_run[proc_id].save()

    dp = self.simulation_run[proc_id].design_point

    if not bool(msg["success"]) or dp.finished == 1:
      return
  

  def _handleNewMetric(self,msg):
    """Method called in the beggining of simulation to register all output metrics that it will spit"""

    print("NEW METRIC received: " + str(msg))

    proc_id = int(msg['id'])
    print("RESULT SET SIMULATION ID = " + str(self.simulation_run[proc_id].id))
    try:
      #Search for an existing metric for this exp. desc.
      #With the same name
      metric = Metric.objects.filter(exp_description_id__exact = self.factory.experiment.exp_description.id).get(name__exact = msg['name'])

      result_set = ResultSet.objects.get(metric = metric, simulation_run = self.simulation_run[proc_id])
      print("Result set Exists")
    except Metric.DoesNotExist:
      #Create object
      metric = Metric()
      metric.exp_description_id = self.factory.experiment.exp_description.id
      metric.name = msg['name']
      metric.data_type = msg['metric_type']
      metric.units = msg['unit']

      #Save in the database
      metric.save()
      if self.pos_offset is None or metric.id < self.pos_offset:
        self.pos_offset = metric.id
      #Do the same for result_set
      result_set = ResultSet()
      result_set.metric = metric
      result_set.simulation_run = self.simulation_run[proc_id]
      result_set.save()
      self.prev_count[int(result_set.id)] = 0
    except ResultSet.DoesNotExist:

      print(int(self.simulation_run[proc_id].id))
      result_set = ResultSet()
      result_set.metric = metric
      result_set.simulation_run = self.simulation_run[proc_id]
      result_set.save()
      self.prev_count[int(result_set.id)] = 0
    except:
      raise Exception("Database in an inconsistent state")

    #Now we should have a consistent metric object
    self.result_sets[str(metric.id) + "_" + str(self.simulation_run[proc_id].id)] = result_set

  def _handleResult(self,msg):
    """Method invoked when a result is received from the worker."""
    proc_id = int(msg['id'])
    simrun = self.simulation_run[proc_id]
    #prevents simulation from processing queued data from the tcp
    #connection ater we've reached the necessary precision
    #and killed the simulator [greatly speeds up simulations of higher precision]
    if simrun.design_point.id != self.dp:
      return
    for data in msg['data']:
      #Get the corresponding metric id (done this way because safe_aggregator stores metrics
      #in an array, and pos is the position of that metric in the array)
      pos = int(data['pos']) + self.pos_offset

      result_set = self.result_sets.get(str(pos) + "_" + str(simrun.id), None)

      if result_set == None:
        result_set = ResultSet()
        result_set.metric = Metric.objects.get(pk = pos)
        result_set.simulation_run = simrun
        result_set.save()
        self.result_sets[str(pos) + "_" + str(simrun.id)] = result_set
        self.prev_count[int(result_set.id)] = 0
      if "value" not in data.keys():
        raise Exception("No value key in the result message")
      value = data["value"]
      time = data["time"]

      binary_time = base64.decodestring(time)
      (actual_time,) = struct.unpack("d", binary_time)

      #Demux msg according to its type
      typeDemux = {
            "double":self._handleDoubleResult,
            "integer":self._handleIntegerResult,
            "boolean":self._handleBooleanResult
          }

      if result_set.metric.data_type in typeDemux.keys():
        typeDemux[result_set.metric.data_type](result_set, actual_time, value, proc_id)
      else:
        raise Exception("Metric type not recognized.")

  def _handleDoubleResult(self, result_set, time, value, proc_id):
    """Method invoked when a double result is received"""

    #First we need to decode from base64 to actual binary
    binary_value = base64.decodestring(value)

    #Now read the binary as a C double (python float)
    (actual_value,) = struct.unpack("d",binary_value)

    #Save it to the database
    result = DoubleData()
    result.type = 'R'
    result.code = result_set.id
    result.value = actual_value
    result.time = time

    self.double_result_buffer.append(result)
    if len(self.double_result_buffer) >= 1000:
      n1 = self.prev_count.get(int(result_set.id), 0)
      self._flushBuffers()
      results = result_set.getResults()
      n2 = results.count()
      self.prev_count[result_set.id]  = n2
      #only works as expeted if n2 - n1 < 1000, which should usually be the case
      if (n1 % 1000) > (n2 % 1000):
        mean = results.aggregate(Avg('value'))['value__avg']
        data = {"type":"estimate", "mean": mean, "results": n2, "metric": result_set.metric.id, "dp_id": result_set.simulation_run.design_point.id}
        self.client.connectedProtocol.sendString(json.dumps(data))
      self._updateDP()

  def _handleIntegerResult(self, result_set, time, value, proc_id):
    """Method invoked when a double result is received"""

    #First we need to decode from base64 to actual binary
    binary_value = base64.decodestring(value)

    #Now read the binary as a C int (python int)
    (actual_value,) = struct.unpack("i",binary_value)

    #Save it to the database
    result = IntegerData()
    result.type = 'R'
    result.code = result_set.id
    result.value = actual_value
    result.time = time

    self.integer_result_buffer.append(result)
    if len(self.integer_result_buffer) >= 1000:
      n1 = self.prev_count.get(int(result_set.id), 0)
      self._flushBuffers()
      results = result_set.getResults()
      n2 = results.count()
      self.prev_count[result_set.id]  = n2
      #only works as expeted if n2 - n1 < 1000, which should usually be the case
      if (n1 % 1000) > (n2 % 1000):
        mean = results.aggregate(Avg('value'))['value__avg']
        data = {"type":"estimate", "mean": mean, "results": n2, "metric": result_set.metric.id, "dp_id": result_set.simulation_run.design_point.id}
        self.client.connectedProtocol.sendString(json.dumps(data))
      self._updateDP()
      
  def _handleBooleanResult(self, result_set, time, value, proc_id):
    """Method invoked when a double result is received"""

    #First we need to decode from base64 to actual binary
    binary_value = base64.decodestring(value)

    (actual_value,) = struct.unpack("?",binary_value)

    #Save it to the database
    result = BooleanData()
    result.type = 'R'
    result.code = result_set.id
    result.value = actual_value
    result.time = time
    
    self.boolean_result_buffer.append(result)
    if len(self.boolean_result_buffer) >= 1000:
      n1 = self.prev_count.get(int(result_set.id), 0)
      self._flushBuffers()
      results = result_set.getResults()
      n2 = results.count()
      self.prev_count[result_set.id]  = n2
      #only works as expeted if n2 - n1 < 1000, which should usually be the case
      if (n1 % 1000) > (n2 % 1000):
        mean = results.aggregate(Avg('value'))['value__avg']
        data = {"type":"estimate", "mean": mean, "results": n2, "metric": result_set.metric.id, "dp_id": result_set.simulation_run.design_point.id}
        self.client.connectedProtocol.sendString(json.dumps(data))
      self._updateDP()

  def _updateDP(self):
    if self.done:
      return
    dp = self.factory.experiment.getCurrentDesignPoint()
    if dp is None and self.dp > -1:
      self.sendString(json.dumps({"type": "kill_dp", 'dp_id':self.dp}))
      self.dp = -1
    elif dp.id > self.dp:
      self.sendString(json.dumps({"type": "kill_dp", 'dp_id':self.dp}))
      self.dp = dp.id
      self.prev_count = {}


class ExperimentFactory(Factory):
  """Class which generates specific protocols objects for each connection.

  The :class:`ExperimentFactory` is responsible for creating all of the
  :class:`ExperimentManager` instances. This is done in the
  :func:`buildProtocol` method which has been overridden from the
  :class:`Factory` superclass.
  """
  def __init__(self, experiment):
    """Initialization of the :class:`ExperimentFactory`.

    This method receives a experiment object from the database that has already a list of design points that will be distributed to workers."""
    self.experiment = experiment

  def buildProtocol(self,addr):
    """Build the :class:`ExperimentManager` protocol.

    This method overrides the default :func:`buildProtocol` method and must
    return a protocol subclass."""
    print("Building a protocol")
    return ExperimentManager(self)

#######################################################
#####             STARTING FUNCTIONS               ####
#######################################################

#Blocking functions to be executed in another thread
#(so it does not block the server)
def startWorkers(experiment, port):
  from django.db import connection
  connection.close()
  t1 = time.time()
  print("START TIME: " + str(t1))
  execute(fabfile.runWorkers, experiment, port)

  reactor.stop()
  t2 = time.time()
  print("FINAL END TIME: " + str(t2))
  print("TOTAL TIME: " + str(t2 - t1))


#Called when fabrics return some error
def startingError(err):
  sys.stderr.write("ERROR WHILE STARTING WORKERS\n")
  sys.stderr.write(str(err) + '\n')
  err.printTraceback()
  reactor.stop()
  sys.exit(1)

def anyError(err):
  print("ANY ERROR")

if __name__ == '__main__':
  #Get id from command line
  parser = argparse.ArgumentParser(description = 'Manage the distributed execution of one experiment')
  parser.add_argument("experiment", help = "Experiment ID to manage execution", type = int)
  #might be best to set this to the number of workers, though we should experiment to find out what works best
  parser.add_argument("--min_results", help = "minimim number of estimates before checking precision", type = int, default = 4)
  parser.add_argument("--conf_level", help = "confidence level to use when calculating precision", type = float, default = 0.90)
  parser.add_argument("--precision", help = "maximum relative error before where data is still considered precise", type = float, default = 0.05)
  parser.add_argument("--server_port", help = "port number that server should run on", type = int, default = 8123)
  parser.add_argument("--analyzer_port", help = "port number that analyzer should run on", type = int, default = None)
  parser.add_argument("--manual-start", dest = 'manual_start', help = "Do not automatically execute workers",action = 'store_true')
  parser.set_defaults(fast_start = False)
  args = parser.parse_args()

  if args.analyzer_port == None:
    args.analyzer_port = args.server_port + 1
  #Get experiment from database
  try:    
    experiment = Experiment.objects.get(pk = args.experiment)
    reactor.listenTCP(args.server_port, ExperimentFactory(experiment))

    #start the global analyzer
    reactor.listenTCP(args.analyzer_port, AnalysisFactory(experiment, args.conf_level, args.precision, args.min_results))

    #Starting workers in a new thread
    #so it doesnot block the server
    if not args.manual_start:
      d = threads.deferToThread(startWorkers,experiment, args.server_port)
      #In case of some error starting workers, stop server
      d.addErrback(startingError)

    reactor.run()

  except ConnectionDone:
    print("teste")

  except Experiment.DoesNotExist, Experiment.MultipleObjectsReturned:
    #In this case, couldnt find experiment in database
    sys.stderr.write("ERROR: No experiment found\n")
    sys.exit(1)
