class AnalysisException(Exception):
        pass

class Analysis:
    """
    Module providing a variety of analysis functions.
    All methods Assumes data is a dictionary where key is x-axis value of data.
    """
        
    def __init__(self, data):
        self.data = data

    def avg_data(self, cluster_size, extract_values, defaults = {}):
        """                                                               
        Generates a new list of data containing the average of all values 
        (from extract_values) from data points within a cluster, where cluster_size
        is the maximum range of values within a cluster. Data should be a dictionary
        of the format <filter_value>:{data}                               
        """
        output = {}
        default = { value:0 for value in extract_values }
        default['size'] = 0
        for point in self.data:
            data = self.data[point]
            index = int(float(point) // cluster_size)
            out_point = output.get(index, dict(default))
            for key in extract_values:
                out_point[key] += data[key]
            out_point['size'] += 1
            output[index] = out_point

        for point in output:
            data = output[point]
            for value in extract_values:
                data[value] /= data['size']
            data.pop('size')
            data.update( defaults )
        return output


    def sliding_window_avg(self, window_size):
        """ 
        Returns the the moving average of the data.
        Each data point is averaged with the window_size points in front of it.
        """ 
        output = {} 
        s_keys = sorted(self.data.keys())
       
        for i in range(len(s_keys) - window_size + 1):
            value = sum([self.data[x]['value'] for x in s_keys[i:i + window_size]])/window_size
            point = self.data[s_keys[i]].copy()
            point['value'] = value
            output[point['time']] = point

        return output
        
               
    def get_cluster_size(self, request):
        """
        Given a request object, returns a cluster_size generated either directly
        from the cluster_size query parameter, or calculated from existing data
        and the max_results query parameter. If either query parameter is less 
        than or equal to zero it raises an exception. If niether parameter exists, returns None
        """
        cluster = request.GET.get('cluster_size',None)
        if cluster is not None and float(cluster) != 0:
            cluster = float(cluster)

            if cluster < 0:
                raise AnalysisException('Bad cluster_size. Cluster size must be an int or float greater than 0.')
            return cluster
        
        max_results = request.GET.get('max_results',None)
        if max_results is not None:
            max_results = int(max_results)

            if max_results < 0:
                raise AnalysisException('Bad max_results. Max Results must be an int greater than 0.')
            if max_results == 0:
                return None
            return (max([float(point) for point in self.data.keys()])/max_results)*1.01
            
        return None        

    def filter_results(self, skip_size):
        """
        Filters via one of two methods, depending on the size of the skip_size.
        If positive, include only every Nth result (e.g. if skip_size is 3, include only every
        third result). If negative filter out every Nth result (e.g. if skip_size is -3, remove every
        third element from the list.
        """
        
        out = {}
        size = abs(skip_size)
        if size == 0:
            return self.data
        
        for i,x in enumerate(sorted(self.data.keys())):
            value = self.data[x]
            if i % size == 0 and skip_size > 0:
                out[value['time']] = value
            elif i % size != 0 and skip_size < 0:
                out[value['time']] = value
        return out

    def filter_outliers(self, mean, std_dev, num_stddevs = 1, filter_value ='value'):
        """
        Given the mean and standard deviation of a set of data, attenot to remove outliers by 
        filtering out values that are more than num_stddevs standard deviations away from the mean.
        """
        if num_stddevs <= 0:
            return self.data
        out = {}
        for key in self.data:
            data = self.data[key]
            value = data[filter_value]
            if abs(value - mean) < std_dev * num_stddevs:
                out[key] = data
        return out

    def filter_all(self, request, defaults = {}):
        """
        Wrapper for the basic filter functions (filter_step, denoise window, and avg_data).
        Requires the request object and optionally default properties to include with every result.
        """
        filter_step = int(request.GET.get('filter_step', 0))
        if filter_step != 0:
           self.data = self.filter_results(filter_step) 
    
        window_size = int(request.GET.get('denoise_window', 0))
        if window_size > 1:
            self.data = self.sliding_window_avg(window_size)     
            
        cluster_size = self.get_cluster_size(request)
        if cluster_size is not None and cluster_size > 0:
            self.data = self.avg_data(cluster_size, ['time','value'], defaults)
            
        return self.data
