from django.db import models
from config import settings
from django.contrib.auth.models import User
from model_utils.managers import InheritanceManager
# Create your models here.

"""
Experiment definition, used to set up design points
"""

class Installation(models.Model):
    name = models.CharField(max_length = 100)
    description = models.CharField(max_length = 255)
    tarball_path = models.FileField(upload_to='installations/')
    simulator_folder = models.CharField(max_length=50)
    public = models.BooleanField(default = False)
    owner = models.ForeignKey(User)


    class Meta:
        unique_together = ('name', 'owner')
    
    def __unicode__(self):
        return self.name

class ExpDescription(models.Model):
    name = models.CharField(max_length = 50)
    description = models.CharField(max_length = 255)
    script = models.CharField(max_length=50)
    installation = models.ForeignKey(Installation)
    owner = models.ForeignKey(User)
    public = models.BooleanField(default=False)

    class Meta:
        unique_together = ('name', 'owner')

class Experiment(models.Model):
    name = models.CharField(max_length=100)
    description = models.CharField(max_length = 255)
    create_date = models.DateTimeField('date created')
    replications = models.IntegerField()
    exp_description = models.ForeignKey(ExpDescription)
    owner = models.ForeignKey(User)
    public = models.BooleanField(default = False)
    
    def __unicode__(self):
        return self.name
    
    def getCurrentDesignPoint(self):
        """Get the designpoint for this experiment with the lowest id that is not finished yet"""
        design_points = self.designpoint_set.filter(finished=0).order_by('pk')
        if design_points.count():
            return list(design_points[:1])[0]        
        return None
    
    class Meta:
        unique_together = ('name', 'owner')

class Metric(models.Model):
    name = models.CharField(max_length = 50)
    exp_description = models.ForeignKey(ExpDescription)
    data_type = models.CharField(max_length = 20, choices=settings.SAFE_TYPES.items(), default = 'double')
    units = models.CharField(max_length = 20)

class Factor(models.Model):
    name = models.CharField(max_length=40)
    description = models.CharField(max_length = 255)
    visible = models.BooleanField(default = True)
    exp_description = models.ForeignKey(ExpDescription)
    objects=InheritanceManager()
    data_type = models.CharField(max_length = 20, choices=settings.SAFE_TYPES.items(), default = 'double')
    units = models.CharField(max_length = 20, default="")
    
    def __unicode__(self):
        return self.name

    def getDefault(self):
        return eval(settings.SAFE_TYPES[self.data_type]).objects.get(type__exact='F', code__exact=self.id).value

   
"""
Pieces of each simulation run
"""

class DesignPoint(models.Model):
    experiment = models.ForeignKey(Experiment)
    finished = models.BooleanField(blank=True, default = 0)

    def countResults(self):
        result_sets = ResultSet.objects.filter(simulation_run__in = self.simulationrun_set.all())
        count = 0
        for result_set in result_sets:
            count += result_set.getResults().count()
        return count
        

            
class Level(models.Model):
    factor = models.ForeignKey(Factor)
    experiment = models.ForeignKey(Experiment)
    design_point = models.ForeignKey(DesignPoint)
    objects=InheritanceManager()
    
    def getValue(self):
        return eval(settings.SAFE_TYPES[self.factor.data_type]).objects.get(type__exact='L', code__exact=self.id).value

class Worker(models.Model):
    path = models.CharField(max_length = 255)
    host = models.CharField(max_length = 255)
    auth_key = models.CharField(max_length = 255, blank=True)
    connected = models.BooleanField(blank=True)
    children = models.IntegerField(default=1)
class SimulationRun(models.Model):
    design_point = models.ForeignKey(DesignPoint)
    start_time = models.DateTimeField('Start time')
    end_time = models.DateTimeField('End time', null=True)
    isFinished = models.BooleanField(default=False)
    success = models.BooleanField(default=False)
    worker = models.ForeignKey(Worker)
    
class ResultSet(models.Model):
    metric = models.ForeignKey(Metric)
    simulation_run = models.ForeignKey(SimulationRun)
    
    def getResults(self):
        #Resutlst of type double are actually stored as double_double where value1 is the value and value2 is the timestamp.
        results = eval(settings.SAFE_TYPES[self.metric.data_type]).objects.filter(type__exact='R', code__exact=self.id).all()
        return results

class GenericData(models.Model):
    #Type can be "R"esult, "F"actor, "L"evel
    type = models.CharField(db_index=True, max_length=1)
    #This is the primary-key of the reuslt,factor or level
    code = models.IntegerField(db_index=True)

    class Meta():
        abstract = True

class IntegerData(GenericData):
    value = models.IntegerField()
    time = models.FloatField(null=True)

class DoubleData(GenericData):
    value = models.FloatField()
    time = models.FloatField(null=True)

class BooleanData(GenericData):
    value = models.BooleanField()
    time = models.FloatField(null=True)

class DoubleDoubleData(GenericData):
    value = models.FloatField()
    value2 = models.FloatField()
    time = models.FloatField(null=True)

class TimestampedDoubleData(GenericData):
    value = models.FloatField()
    time = models.FloatField(null=True)
