#
# Copyright (c) 2013 Bucknell University
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation;
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Authors: William Stratton (wss012@bucknell.edu)
#          Vinicius Felizardo (vdf001@bucknell.edu)
#
from django.http import HttpResponse
from safe.models import *
from django.template import Context, RequestContext
from django.shortcuts import *
from django.core.management import call_command
from django.contrib.auth import authenticate
from django.contrib.auth.decorators import user_passes_test, login_required
from django.utils.decorators import method_decorator
from django.utils import timezone
from django.db.models import StdDev, Avg
from rest_framework import viewsets, generics, status
from rest_framework.permissions import *
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.views import APIView
from safe.serializers import *
from safe.permissions import *
from rest_framework.authtoken.models import Token
from analysis import *
import config.settings
import itertools
import sys
import json
import os
import subprocess

def main(request):
  """ returns the main SAFE webpage"""
  return render_to_response('safe/index.html', context_instance = RequestContext(request))

@api_view(['GET', 'POST'])
@login_required
def run_experiment(request, exp_id):
  if Experiment.objects.filter(pk = exp_id).count() == 1:
    if Worker.objects.filter(connected = 0).count() > 0:
      args = {}
      args['min_results'] = request.GET.get('min_results', None) #minimim number of results before checking precision
      args['conf_level']  = request.GET.get('confidence', None) #confidence level to use when calculating precision
      args['precision']  = request.GET.get('precision', None) #maximum relative error before where data is still considered precise
      args['server_port']  = request.GET.get('precision', None) #port that server process should run on
      args['analyzer_port']  = request.GET.get('precision', None) #port that global analyzer process should run on
      arg_string = ""
      for key,value in args.iteritems():
        if value != None:
          arg_string += ' --' + str(key) + ' ' + str(value)
      EEMpy = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', 'EEM.py')
      subprocess.Popen(EEMpy + " " + str(exp_id) + arg_string, shell = True)

      return HttpResponse("Started")
    else:
      return HttpResponse("All workers are busy")
  else:
    return HttpResponse("Invalid Experiment ID")

@login_required
def delete_experiment(request, exp_id):
  experiment = Experiment.objects.get(pk=exp_id)
  if request.user.is_superuser or request.user == experiment.owner:
    experiment.delete()
    return HttpResponse("OK")
  return HttpResponse("You are not the owner of that experiment")

def dp_query(request,exp_id):
  """ Retrieves a list of design points and their associated
      information (such as factors and levels) """

  dp_set = DesignPoint.objects.filter(experiment = exp_id)
  out = []
  exp_desc_id = Experiment.objects.get(pk = exp_id).exp_description
  factors = Factor.objects.filter(visible = 1, exp_description = exp_desc_id)

  for dp in dp_set:
    data = {'id':dp.id}
    factor_list = {}
    for factor in factors:
      level = Level.objects.get(factor_id = factor.id, design_point_id = dp.id)
      factor_list[factor.name] = level.getValue()
    data['factors'] = factor_list
    out.append(data)

  return HttpResponse(json.dumps(out), content_type = "application/json")

def simrun_dp_query(request, dp_id):
  """ Retrieves the the set of all simulation runs for the given DesignPoint in JSON format  """
  sim_runs = DesignPoint.objects.get(pk = dp_id).simulationrun_set.filter(success = 1)
  out = [{"simrun_id":run.id} for run in sim_runs]

  return HttpResponse(json.dumps(out), content_type = "application/json")

def dp_factor_query(request):
  """ Gets a list of all designpoints whose factor levels match the given values """
  factor_name = request.GET.get('factor','')
  exp_id = request.GET.get('exp_id','')
  level_value = request.GET.get('value','')
  factor = Experiment.objects.get(pk = exp_id).exp_description.factor_set.get(name = factor_name)
  if not factor:
    return Response("Factor Name DOES NOT EXIST", status = status.HTTP_400_BAD_REQUEST)
  levels = factor.level_set.all()
  out = [ {"designpoint_id":level.design_point.id} for level in levels if float(level.getValue()) == float(level_value) ]

  return HttpResponse(json.dumps(out), content_type = "application/json")


def results_query(request, sim_run_id, metric_id):
  """ retreives a set of results in JSON format given the simulation
      run id and the metric id """

  #get results from database
  result_set = ResultSet.objects.get(simulation_run_id = sim_run_id, metric_id = metric_id)
  results = result_set.getResults()

  #convert to analysis-module compatible array
  out = {r.time: {'result_id': r.id, 'time': r.time, 'value': r.value} for r in results}

  analyzer = Analysis(out)

  num_outliers = request.GET.get('filter_outliers', None)
  if num_outliers is not None:
    stats = results.aggregate(StdDev('value'), Avg('value') )
    out = analyzer.filter_outliers(stats['value__avg'], stats['value__stddev'], float(num_outliers))
    analyzer = Analysis(out)

  try:
    out = analyzer.filter_all(request)
  except AnalysisException as e:
    return HttpResponse(str(e), status = status.HTTP_400_BAD_REQUEST)

  out = out.values()
  return HttpResponse(json.dumps(out), content_type = "application/json")


def factor_results(request, exp_id, factor_name, metric_id):
  """Gets a list of results for a given metric for all values of the given factor. Requires that all factors other than the 
  provided factor be listed in the body of the request along with an associated value."""
  decoder = json.JSONDecoder()
  data = decoder.decode(request.body)
  factor_names = data['factors']
  factors = Factor.objects.filter(name__in = factor_names)
  dp_ids = set()
  for factor in factors:
    dps = set()
    value = float(data['factors'][factor.name])
    for l in factor.level_set.filter(experiment_id = exp_id):
      if float(l.getValue()) == float(value):
        dps.add(l.design_point.id)
    if len(dp_ids) == 0:
      dp_ids = dps
    else:
      dp_ids = dp_ids & dps

  design_points = DesignPoint.objects.filter(pk__in = list(dp_ids))
  simruns = SimulationRun.objects.filter(design_point__in = design_points)
  result_sets = ResultSet.objects.filter(simulation_run__in = simruns, metric_id = metric_id)
  factor = Factor.objects.get(name = factor_name)
  #get results from database
  out = []

  analyze = request.GET.get('max_results',False) or request.GET.get('cluster_size',False)  or \
            request.GET.get('denoise_window', False) or request.GET.get('filter_step', False) or \
            request.GET.get('filter_outliers', False)

  for result_set in result_sets:
    results = result_set.getResults()
    simrun = result_set.simulation_run
    s_id = simrun.id
    level_value = simrun.design_point.level_set.get(factor = factor).getValue()
    if analyze:
      temp = {}
      for r in results:
        temp[r.time] = {'time': r.time, 'value': r.value}

      analyzer = Analysis(temp)
      num_outliers = request.GET.get('filter_outliers', None)
      if num_outliers is not None:
        stats = results.aggregate(StdDev('value'), Avg('value') )
        temp = analyzer.filter_outliers(stats['value__avg'], stats['value__stddev'], float(num_outliers))
        analyzer = Analysis(temp)

      try:
        temp = analyzer.filter_all(request, {'level':level_value})
      except AnalysisException as e:
        return HttpResponse(str(e), status = status.HTTP_400_BAD_REQUEST)
      out.extend(temp.values())

    else:
     for r in results:
        out.append({'time': r.time, 'value': r.value, 'level': level_value})

  return HttpResponse(json.dumps(out), content_type = "application/json")

def multi_results_query(request, metric_id):
  """ retreives a set of all results in JSON format given the metric id in the url
      and a list of simrun_id's in the body in JSON format """
  decoder = json.JSONDecoder()
  data = decoder.decode(request.body)
  srun_ids = [ int(s['simrun_id']) for s in data ]
  #get results from database
  result_sets = ResultSet.objects.filter(simulation_run_id__in = srun_ids, metric_id = metric_id)
  out = []

  analyze = request.GET.get('max_results',False) or request.GET.get('cluster_size',False)  or \
            request.GET.get('denoise_window', False) or request.GET.get('filter_step', False) or \
            request.GET.get('filter_outliers', False)

  for result_set in result_sets:
    results = result_set.getResults()
    simrun = result_set.simulation_run
    s_id = simrun.id
    d_id = simrun.design_point.id
    if analyze:
      temp = {}
      for r in results:
        temp[r.time] = {'time': r.time, 'value': r.value}

      analyzer = Analysis(temp)
      num_outliers = request.GET.get('filter_outliers', None)
      if num_outliers is not None:
        stats = results.aggregate(StdDev('value'), Avg('value') )
        temp = analyzer.filter_outliers(stats['value__avg'], stats['value__stddev'], float(num_outliers))
        analyzer = Analysis(temp)

      try:
        temp = analyzer.filter_all(request, {'simrun_id': s_id, 'designpoint_id':d_id})
      except AnalysisException as e:
        return HttpResponse(str(e), status = status.HTTP_400_BAD_REQUEST)
      out.extend(temp.values())

    else:
     for r in results:
        out.append({'time': r.time, 'value': r.value, 'simrun_id': s_id, 'designpoint_id': d_id})

  return HttpResponse(json.dumps(out), content_type = "application/json")


def aggregate_query(request, dp_id, metric_id):
  """ Aggregates all replications of a single design point
      for a single metric, returning the "average" data set """
  simruns = DesignPoint.objects.get(pk = dp_id).simulationrun_set.filter(finished = 1)
  #s_ids = [int(s.id) for s in simruns]
  result_sets = ResultSet.objects.filter(simulation_run__in = simruns, metric_id = int(metric_id))
  if (len(result_sets) < 1):
    return HttpResponse(json.dumps([]), content_type = "application/json")

  max_avg = 0
  avg_size = 0
  out = {}
  agg_results = None
  num_outliers = request.GET.get('filter_outliers', None)

  for result_set in result_sets:
    max_time = 0
    results = result_set.getResults()
    if num_outliers is not None:
      if agg_results is None:
        agg_results = results
      else:
        agg_results = agg_results | results
    avg_size += len(results)
    for result in results:
      if result.time > max_time:
        max_time = result.time
      out[result.time] = {'value':result.value, 'time':result.time}
    max_avg += max_time

  avg_size /= len(result_sets)
  max_avg /= len(result_sets)
  analyzer = Analysis(out)
  if num_outliers is not None:
    stats = agg_results.aggregate(StdDev('value'), Avg('value') )
    out = analyzer.filter_outliers(stats['value__avg'], stats['value__stddev'], float(num_outliers))
    analyzer = Analysis(out)
  out = analyzer.avg_data(max_avg / avg_size, ['time','value'])
  analyzer = Analysis(out)

  try:
    out = analyzer.filter_all(request)
  except AnalysisException as e:
    out = out.values()
  return HttpResponse(json.dumps(out), content_type = "application/json")

def metrics_query(request, exp_id):
  """ returns all of the metrics associated with a given simulation run"""

  #get metrics
  met = Experiment.objects.get(pk = exp_id).exp_description.metric_set.all()
  #met = SimulationRun.objects.get(pk = sim_run_id).design_point.experiment.exp_description.metric_set.all()

  #convert to JSON
  out = [{'id': str(m.id), 'name': m.name, 'units': m.units, 'type': m.data_type} for m in met]

  return HttpResponse(json.dumps(out), content_type = "application/json")


def dp_range_query(request, exp_id):
  """ Returns a list of all designpoints as JSON with levels between the
   min and max values specified in the body as JSON """
  decoder = json.JSONDecoder()
  data = decoder.decode(request.body)
  d = DesignPoint.objects.filter(experiment_id = exp_id)
  points = {int(x.id) for x in d}
  for l in data:
    factor = Factor.objects.get(name = l['name'])
    levels = factor.level_set.all()
    max_val = float(l['max_value'])
    min_val = float(l['min_value'])
    dps = { int(level.design_point.id) for level in levels if min_val <= level.getValue() <= max_val }
    points = points.intersection(dps)
    if len(points) == 0:
      break
  return HttpResponse(json.dumps(list(points)), content_type = "application/json")

def data_query(request, sim_run):
  """ Returns a list of all results for the given simulation run in JSON format..
      Data points are returned by time, and each point contains the value
      for all metrics specified in the query string or all metrics if not specified """
  result_sets = SimulationRun.objects.get(pk = sim_run).resultset_set
  #consider changing this to use the body of the request, rather than a query parameter
  if request.GET.get('metrics', None) == None :
   result_sets = result_sets.all()
  else:
    metrics = [int(x) for x in request.GET.get('metrics', None).split(',')]
    result_sets = result_sets.filter(metric_id__in = metrics)
  results = {}
  for r_set in result_sets:
    rows = r_set.getResults()
    for data in rows:
      results[data.time] = results.get(data.time, {'time':data.time})
      results[data.time].update({'metric_' + str(r_set.metric_id):data.value})
  return HttpResponse(json.dumps(results), content_type = 'application/json')


def desc_experiments(request, user_id):
  """Returns a list of all experiments belonging to the given user in JSON format.
  """
  if str(request.GET.get('public', '0')) == '1' :
    experiments = Experiment.objects.filter(owner_id = user_id) | Experiment.objects.filter(public = 1)
  else:
    experiment = Experiment.objects.filter(owner_id = user_id)
  out = [{'exp_id':experiment.id} for experiment in experiments]
  return HttpResponse(json.dumps(out), content_type='application/json')

def desc_expdesc(request, user_id):
  """Returns a list of all experiments belonging to the given user in JSON format.
  """
  if str(request.GET.get('public', '0')) == '1' :
    experiments = ExpDescription.objects.filter(owner_id = user_id) | ExpDescription.objects.filter(public = 1)
  else:
    experiments = ExpDescription.objects.filter(owner_id = user_id)
    
  out = [{'expdesc_id':experiment.id} for experiment in experiments]

  return HttpResponse(json.dumps(out), content_type = 'application/json')

@api_view(['GET'])
def desc_current_experiments(request):
  """Returns a list of all experiments belonging to the current user in JSON format.
  """

  if not request.user.is_authenticated():
    return HttpResponse("User must be logged in to retrieve current experiments", status = status.HTTP_403_FORBIDDEN)

  experiments = Experiment.objects.filter(owner = request.user)

  out = [{'exp_id':experiment.id, 'exp_name':experiment.name} for experiment in experiments]

  return HttpResponse(json.dumps(out), content_type = 'application/json')

def exp_status(request, exp_id):
  """ Gets the status of the current experiment """
  exp = Experiment.objects.get(pk = exp_id)
  dpts = exp.designpoint_set.all()
  out = {'dpts': len(dpts), 'replications': exp.replications}
  simruns = SimulationRun.objects.filter(design_point__in = dpts)
  errors = []
  complete = 0
  for s in simruns:
    complete += s.isFinished
    if s.isFinished and not s.success:
      errors.append({'dpt': s.design_point_id, 'simrun': s.id})

  out['errors'] = errors
  out['complete'] = complete
  return HttpResponse(json.dumps(out), content_type = 'application/json')

def exp_simruns(request, exp_id):
  """ Returns a list of all simulation runs and their associated design point
  for a given simulation run in JSON format for a given experiment """
  design_points = Experiment.objects.get(pk = exp_id).designpoint_set.all()
  simruns = []
  for dp in design_points:
    simruns.extend(dp.simulationrun_set.all())
  out = []
  for simrun in simruns:
    out.append({"id":simrun.id, "dp_id":simrun.design_point.id})
  return HttpResponse(json.dumps(out), content_type = 'application/json')

def exp_levels(request, exp_id):
  """returns a list of all factors and the levels available for each factor in JSON Format """
  factors = Experiment.objects.get(pk = exp_id).exp_description.factor_set.all()
  out = {}
  for f in factors:
    data = {"name":f.name,"description":f.description,"id":f.id, "levels":set()}
    for l in f.level_set.filter(experiment_id = exp_id):
        data['levels'].add(l.getValue())
    data['levels'] = list(data['levels'])
    out[f.id] = data
  return HttpResponse(json.dumps(out,sort_keys=True), content_type='application/json')

def get_exp_info(request, exp_id):
  """returns a variety of information pertaining to an experiment"""
  experiment = Experiment.objects.get(pk = exp_id)
  out = {}
  out['name'] = experiment.name
  out['replications'] = experiment.replications
  out['description'] = experiment.description
  out['creation_date'] = str(experiment.create_date)
  out['exp_description_id'] = experiment.exp_description.id
  out['owner'] = experiment.owner.id
  out['public'] = experiment.public
  dp_set = DesignPoint.objects.filter(experiment = exp_id)
  out['design_points'] = [dp.id for dp in dp_set]

  return HttpResponse(json.dumps(out,sort_keys = True), content_type = 'application/json')

@api_view(['POST'])
@user_passes_test(lambda u: u.has_perm('add_expdescription'))
def sled_parser(request):
    """ Parses a SLED json file aand uses it to set up an experiment description """
    decoder = json.JSONDecoder()
    data = decoder.decode(request.body)
    if data['owner'] == None:
      data['owner'] = request.user.id
    else:
      data['owner'] = User.objects.get(username = data['owner']).id

    try:
      installation = Installation.objects.get(owner_id = data['owner'], name = data['installation'])
    except Installation.DoesNotExist:
        return Response("DOES NOT EXIST", status = status.HTTP_400_BAD_REQUEST)

    #Creating experiment description
    exp_desc = ExpDescription()
    exp_desc.installation = installation
    exp_desc.name = data['name']
    exp_desc.description = data['description']
    exp_desc.script = data['script']
    exp_desc.public = data['public']
    exp_desc.owner = request.user
    exp_desc.save()

    #Creating factors
    for data_factor in data['factors']:
        factor = Factor()
        factor.name = data_factor['name']
        factor.visible = data_factor['visible']
        factor.description = data_factor['description']
        factor.data_type = data_factor['type']
        factor.exp_description = exp_desc
        factor.units = data_factor['units']
        factor.save()

        default_data = eval(settings.SAFE_TYPES[data_factor['type']])()
        default_data.type = 'F'
        default_data.code = factor.id
        default_data.value = data_factor['default']
        default_data.save()

    #DONE
    return Response(json.dumps({'Description id':exp_desc.id}), status = status.HTTP_201_CREATED, content_type = 'application/json')




def exp_factors(request, exp_id):
    """ returns a list of all factors available for a given experiment in JSON format"""
    exp_desc = Experiment.objects.get(pk = exp_id).exp_description
    factors = exp_desc.factor_set.filter(visible = True)

    fact_list = [{'name': f.name, 'description': f.description, 'type': f.data_type, 'id': f.id} for f in factors]

    return HttpResponse(json.dumps(fact_list), content_type = 'application/json')


def desc_factors(request, desc_id):
    """ returns a list of all factors for a given experiment description in JSON format"""
    exp_desc = ExpDescription.objects.get(pk = desc_id)
    factors = exp_desc.factor_set.filter(visible = True)

    fact_list = [{'name': f.name, 'description': f.description, 'type': f.data_type, 'id': f.id} for f in factors]

    return HttpResponse(json.dumps(fact_list), content_type = 'application/json')

@api_view(['POST'])
@login_required
@user_passes_test(lambda u: u.has_perm('add_experiment'))
def level_upload(request):
  """
  Receives an experiment configuration file via POST, then
  stores the levels and creates design points from the levels.
  """
  decoder = json.JSONDecoder()
  data = decoder.decode(request.body)

  factor_ids = []
  level_sets = []
  for factor in data['factors']:
    fact = Factor.objects.get(exp_description_id = data['exp_description_id'], name = factor['name'])
    f_id = fact.id
    factor_ids.append(f_id)
    if factor.get('levels',None) is None or len(factor['levels']) == 0:
      level_sets.append([fact.getDefault()])
    level_sets.append(factor['levels'])

  #remove factors from data so we can use data to initialize the experiment
  data.pop('factors')

  if data["public"] in (1, "1", True):
    data["public"] = True 
  else:
    data["public"] = False 
  experiment = Experiment(**data)
  experiment.owner = request.user
  experiment.create_date = timezone.now()
  experiment.save()
  print(level_sets)

  design_points = list(itertools.product(*level_sets))
  print(design_points)
  for d_pt in design_points:
    d = DesignPoint(experiment_id = experiment.id)
    d.save()
    print(d)

    for i in range(len(d_pt)):
      l = Level()
      l.factor_id = factor_ids[i]
      l.experiment_id = experiment.id
      l.design_point_id = d.id
      l.save()

      g = eval(settings.SAFE_TYPES[Factor.objects.get(pk = l.factor_id).data_type])()
      g.type = 'L'
      g.code = l.id
      g.value = d_pt[i]
      g.save()


  return Response("OK", status = status.HTTP_201_CREATED)

@api_view(['POST'])
#@user_passes_test(lambda u: u.has_perm('add_user'))
def create_user(request):
  """
  Creates a user given data on each item.
  """

  data = request.DATA

  # Add the basic information
  user = User.objects.create_user(username = data['username'], email = data['email'], password = data['password'], first_name = data['first_name'], last_name = data['last_name'])
  # Set permissions
  user.is_superuser = data['is_superuser']
  user.is_staff = data['is_staff']
  # Hashes the password
  user.set_password(data['password'])
  user.save()

  return Response("OK", status = status.HTTP_201_CREATED)

################################################################
#
# Code below this point creates view sets for each piece of the model
#
################################################################

class InstallationViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows installations to be viewed or edited.
  """
  queryset = Installation.objects.all()
  serializer_class = InstallationSerializer
  permission_classes = (IsOwnerOrPublic,DjangoModelPermissionsOrAnonReadOnly,)

  def pre_save(self, obj):
    """
    Set the object's owner as the person who created
    """
    obj.owner = self.request.user

class ExperimentViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows experiments to be viewed or edited.
  """
  queryset = Experiment.objects.all()
  serializer_class = ExperimentSerializer
  permission_classes = (IsOwnerOrPublic,IsAuthenticatedOrReadOnly,)

  def pre_save(self, obj):
    """
    Set the object's owner as the person who created
    """
    obj.owner = self.request.user
    obj.create_date = timezone.now()

class ExpDescViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows experiment descriptions to be viewed or edited.
  """
  queryset = ExpDescription.objects.all()
  serializer_class = ExpDescriptionSerializer
  permission_classes = (IsOwnerOrPublic,IsAuthenticatedOrReadOnly,)

  def pre_save(self, obj):
    """
    Set the object's owner as the person who created
    """
    obj.owner = self.request.user

class UserViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows users to be viewed or editied.
  """
  queryset = User.objects.all()
  serializer_class = UserSerializer

  #Automatically create token for the user
  def post_save(self, instance = None, created = False):
    if created:
      Token.objects.create(user = instance)

class MetricViewSet(viewsets.ReadOnlyModelViewSet):
  """
  API endpoint that allows models to be viewed.
  """
  queryset = Metric.objects.all()
  serializer_class = MetricSerializer

class FactorViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows factors to be viewed or edited.
  """
  queryset = Factor.objects.all()
  serializer_class = FactorSerializer

class WorkerViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows workers to be viewed or edited.
  """
  queryset = Worker.objects.all()
  serializer_class = WorkerSerializer

class DesignPointViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows design points to be viewed and edited.
  """
  queryset = DesignPoint.objects.all()
  serializer_class = DesignPointSerializer

class SimulationRunViewSet(viewsets.ModelViewSet):
  """
  API endpoint that allows simulation runs to be viewed and edited.
  """
  queryset = SimulationRun.objects.all()
  serializer_class = SimulationRunSerializer
