/**-------------------------------------
 * The base cartesian graph object
 * -------------------------------------
 *
 * This is the base carteisna graph object. It is expected that all
 * graphs will make use of this object's initialization and
 * ptototype for their functionality.
 */

safe.graph.cartesian = safe_graph_cartesian;
safe.graph.clipPathNumber = 0;

safe_graph_cartesian.prototype = Object.create(safe.graph.base.prototype);
safe_graph_cartesian.constructor = safe_graph_cartesian;

/**
 * Sets up the layout of the graph and adds it to the specified element.
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 *
 * @return The newly created object
 * */

function safe_graph_cartesian(margins, width, height, selectedCallBack, parentElement){
    safe.graph.base.apply(this, [margins, width, height, selectedCallBack, parentElement]);
    this.xMin = this.yMin = this.xMax = this.yMax = 0;
    this.data = {};
    this.xAccessor = {};
    this.yAccessor = {};
    this.toDisplay = {};
    this.pointAttributes = {};

    this.tooltip = d3.select("body").append("div")
        .attr("class", "tooltip")
        .style("opacity", 0);

    return this;
}

/** Redisplays all data sets for this graph */
safe_graph_cartesian.prototype.displayAll = function(){
    Object.keys(this.data).forEach(function(id){
        if(this.toDisplay[id] == true) { this.display(id) }
    }.bind(this));
}

/** 
 * Adds the specified data set to this graph
 * @param id A string that distinguishes this data set from all others on this graph
 * @param data The data to be added to this graph
 * @param xAccessor A function which will return the x value of a point when called with the point as a parameter
 * @param yAccessor A function which will return the y value of a point when called with the point as a parameter
 * @param pointAttributes A dictionary of html attributes to add to each displayed point
 * @param mouseEvents A dictionary of mouse events to occur when mouse is over a data point
 */
safe_graph_cartesian.prototype.addDataSet = function(id, data, xAccessor, yAccessor, pointAttributes){
    this.data[id] = data;
    this.xAccessor[id] = xAccessor;
    this.yAccessor[id] = yAccessor;
    this.pointAttributes[id] = pointAttributes;
    this.toDisplay[id] = true;
    this.updateDataFeatures();
    this.displayAll();
}

/** 
 * Sets the data for the specified data set 
 * @param id A string that distinguishes this data set from all others on this graph
 * @param data The data to be added to this graph
 */
safe_graph_cartesian.prototype.addData = function(id, data){
    this.data[id] = data;
}

/**
 * Sets the x value accessor function for the specified data set
 * @param id The string id for the desired data set specified in "addDataSet"
 * @param xAccessor A function that will return the x value of a point when called with the point as a parameter
 */
safe_graph_cartesian.prototype.setXAccessor = function(id, xAccessor){
    this.xAccessor[id] = xAccessor;
}

/**
 * Sets the y value accessor function for the specified data set
 * @param id The string id for the desired data set specified in "addDataSet"
 * @param yAccessor A function that will return the y value of a point when called with the point as a parameter
 */
safe_graph_cartesian.prototype.setYAccessor = function(id, yAccessor){
    this.yAccessor[id] = yAccessor;
}

/** 
 * Returns the data associated with the given id
 * @param id The string id for the desired data set specified in "addDataSet"
 * @return The data associated with the given id
 */
safe_graph_cartesian.prototype.getDataSet = function(id){
    return this.data[id];
}

/** 
 * Returns all data associated with this graph 
 * @return The data associated with this graph
 */
safe_graph_cartesian.prototype.getData = function(){
    results = {};
    Object.keys(this.data).forEach(function(key){
        if(this.toDisplay[key] == true)
            results[key] = this.data[key];
    }.bind(this));
    return results;
}

/**
 * Determines whether the data associated with the given id exists
 * @param id The string id for the desired data set specified in "addDataSet"
 * @return true if the data set exists, false otherwise
 */
safe_graph_cartesian.prototype.hasDataSet = function(id){
    return this.data[id] != undefined;
}

/** 
 * Returns the dictionary of attributes associated with the specified id
 * @param id The string id for the desired data set specified in "addDataSet"
 * @return The dictionary of attributes associated with the specified id
 */
safe_graph_cartesian.prototype.getPointAttributes = function(id){
    return this.pointAttributes[id];
}

/** 
 * Sets the specified data set as invisible without removing its data from memory
 * @paran id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_cartesian.prototype.setInvisible = function(id){
    if(this.toDisplay[id] == false){ return }
    this.toDisplay[id] = false;
    this.updateDataFeatures();
    this.display(id);
    this.displayAll();
}

/** 
 * Sets the specified data set as visible
 * @paran id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_cartesian.prototype.setVisible = function(id){
    if(this.toDisplay[id] == true){ return }
    this.toDisplay[id] = true;
    this.updateDataFeatures();
    this.display(id);
    this.displayAll();
}

/** 
 * Returns a list of all ids currently visible on this graph
 * @return A list of all ids currently visible on this graph
 */
safe_graph_cartesian.prototype.getVisibleIds = function(){
    values = [];
    Object.keys(this.toDisplay).forEach(function(id){
        if(this.toDisplay[id] == true)
            values.push(id);
    }.bind(this));
    return values;
}

/** 
 * Removes all information about the specified data set from the graph.
 * @param id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_cartesian.prototype.remove = function(id){
    this.setInvisible(id);
    delete this.data[id];
    delete this.pointAttributes[id];
}

/** Removes all information for all data sets associated with this graph */
safe_graph_cartesian.prototype.removeAll = function(){
    Object.keys(this.data).forEach(function(id){
        this.setInvisible(id);
        this.remove(id);
    }.bind(this));
}

/** Updates cross hairs on this graph if they are enabled */
safe_graph_cartesian.prototype.updateCrossHairs = function(){
    var x = d3.event.offsetX;
    var y = d3.event.offsetY;

    var xleft = this.xScale.range()[0] + this.margin.left; 
    var xright = this.xScale.range()[1] + this.margin.left;
    var ytop = this.yScale.range()[0] + this.margin.top;
    var ybottom = this.yScale.range()[1] + this.margin.top;

    if (x < xleft) { x = xleft }
    if (x > xright) { x = xright }
    if (y < ybottom) { y = ybottom }
    if (y > ytop) { y = ytop }

    horizontal = this.svg.selectAll(".crosshairhorizontal").data([""]);
    vertical = this.svg.selectAll(".crosshairvertical").data([""]);
 
    horizontal.enter().append("line").attr("class", "crosshairhorizontal").style({"stroke": "black", "pointer-events": "none"})
    vertical.enter().append("line").attr("class", "crosshairvertical").style({"stroke": "black", "pointer-events": "none"})
   
    horizontal.attr("x1", xleft).attr("x2", xright).attr("y1", y).attr("y2", y);
    vertical.attr("y1", ybottom).attr("y2", ytop).attr("x1", x).attr("x2", x);

    horizontal.exit().remove();
    vertical.exit().remove();
}

/** Enables cross hairs for this graph */
safe_graph_cartesian.prototype.enableCrossHairs = function(){
    this.svg.on("mousemove", this.updateCrossHairs.bind(this));
    this.svg.on("mouseout", function(){
        this.svg.selectAll(".crosshairhorizontal").remove();
        this.svg.selectAll(".crosshairvertical").remove();
        }.bind(this));
}

/** Disables cross hairs for this graph */
safe_graph_cartesian.prototype.disableCrossHairs = function(){
    this.svg.selectAll(".crosshairhorizontal").remove();
    this.svg.selectAll(".crosshairvertical").remove();
    this.svg.on("mousemove", undefined);
}

/** Called when mouse is over a data point */
safe_graph_cartesian.prototype.mouseOverPoint = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0.9);

    this.tooltip.html(d3.event.target.getAttribute("info"))
        .style("left", (d3.event.pageX) + "px")
        .style("top", (d3.event.pageY - 28) + "px");
}

/** Called when mouse is no longer over a data point */
safe_graph_cartesian.prototype.mouseOutPoint = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0);
}

/** 
 * Sets the height of the graph (including margins)
 * @param height The pixel height for the graph
 */
safe_graph_cartesian.prototype.setHeight = function(height){
    safe.graph.base.prototype.setHeight.apply(this, [height]);
    this.updateBrush();
}

/** 
 * Sets the width of the graph (including margins)
 * @param width The pixel width for the graph
 */
safe_graph_cartesian.prototype.setWidth = function(width){
    safe.graph.base.prototype.setWidth.apply(this, [width]);
    this.updateBrush();
}

/**
 * Creates a brush to be used on this graph.
 * @param brushCallBack A function to be called when the brush is resized
 *        called as "brushCallBack(brush)"
 * @param constrain An axis constant from safe.constants indicating an axis
 *        to be constrained
 */
safe_graph_cartesian.prototype.addBrush = function(brushCallBack, constrain){
    this.brushConstrain = constrain;
    this.brushCallBack = brushCallBack;
    this.brush = d3.svg.brush().on("brush", this.brushed.bind(this));

    this.svg.append("g")
        .attr("class", "brush")
        .attr("transform", "translate(" + this.margin.left + "," + this.margin.top + ")")
        .call(this.brush);

    this.updateBrush();
}

/**
 * Calls the function associated with this graph's brush
 * if a brush has been created though the addBrush function
 * The function is called as brushCallBack(this.brush)
 */
safe_graph_cartesian.prototype.brushed = function(){
    if(!this.brush){ return }

    this.brushCallBack(this.brush);
}

/**
 * Updates the brush based on the current size of the graph.
 */
safe_graph_cartesian.prototype.updateBrush = function(){
    if(! this.brush){ return }

    if(this.brushConstrain == safe.constants.Y_AXIS){
        this.brush.x(this.xScale);
        this.svg.selectAll(".brush").call(this.brush)
            .selectAll("rect")
            .attr({"height": this.height + 7, "y": -6 });

    }else if(this.brushConstrain == safe.constants.X_AXIS){
        this.brush.y(this.yScale);
        this.svg.selectAll(".brush").call(this.brush)
            .selectAll("rect")
            .attr({"width": this.width, "x": 1 })

    }else{
        this.brush.x(this.xScale).y(this.yScale);
        this.svg.selectAll(".brush").call(this.brush);
    }
}

/** Updates information about maximum and minimum value in the data sets */
safe_graph_cartesian.prototype.updateDataFeatures = function(){
    this.xMin = this.xMax = this.yMin = this.yMax = undefined;
    Object.keys(this.data).forEach(function(id){
        if(this.toDisplay[id] != true){ return }

        xRange = d3.extent(this.data[id], function(d) { return this.xAccessor[id](d) }.bind(this));
        this.xMin = (this.xMin == undefined) ? xRange[0] : d3.min([this.xMin, xRange[0]]);
        this.xMax = (this.xMax == undefined) ? xRange[1] : d3.max([this.xMax, xRange[1]]);

        yRange = d3.extent(this.data[id], function(d) { return this.yAccessor[id](d) }.bind(this));
        this.yMin = (this.yMin == undefined) ? yRange[0] : d3.min([this.yMin, yRange[0]]);
        this.yMax = (this.yMax == undefined) ? yRange[1] : d3.max([this.yMax, yRange[1]]);
        }.bind(this));

    this.setDomains([this.xMin, this.xMax], [this.yMin, this.yMax]);
}
