/**----------------------------------
 * A graph for histograms
 *----------------------------------
 */

safe.graph.barchart = safe_graph_barchart;

safe_graph_barchart.prototype = Object.create(safe.graph.base.prototype);
safe_graph_barchart.constructor = safe_graph_barchart;

/**
 * Creates a new bar chart object
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 *
 * @return The newly created object
 * */
function safe_graph_barchart(margins, width, height, onSelectedCallBack, parentElement){
    safe.graph.base.apply(this, [margins, width, height, onSelectedCallBack, parentElement]);

    // Set x axis as discrete (not continuous as by default)
    this.xScale = d3.scale.ordinal().rangeRoundBands([0, this.width], 0.05);
    this.yAxis.scale(this.xScale);

    this.infoFormat = d3.format(".4f");
    this.tickFormat = d3.format(".2f");
    this.tooltip = d3.select("body").append("div")
        .attr("class", "tooltip")
        .style("opacity", 0);

    return this;
}

/** Displays all data asssociated with this graph */
safe_graph_barchart.prototype.displayAll = function(){
    this.display();
}

/**
 * Displays the specified data set in the form associated with line graphs
 * @param id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_barchart.prototype.display = function(){
    if(this.data == undefined){ return }

    var bars = this.svg.selectAll(".bar").data(this.data);
    bars.enter().append("g").attr("class", "bar");
    bars.exit().remove();

    bars.selectAll("rect").remove();
    bars.append("rect");

    var rects = bars.selectAll("rect");
    rects
        .attr("x", function(d) { return this.margin.left + this.xScale(d.x  + 0.05 * d.dx) }.bind(this))
        .attr("y", function(d) { return this.margin.top + this.yScale(d.y)}.bind(this))
        .attr("width", function(d) { return this.xScale(this.xScale.domain()[0] + d.dx) * 0.9 }.bind(this))
        .attr("height", function(d) { return this.height - this.yScale(d.y) }.bind(this))
        .attr("info", function(d) { return this.infoFormat(d.y) + "</br>" + 
                                    this.infoFormat(d.x) + " - " + this.infoFormat(parseFloat(d.x) + parseFloat(d.dx)) }.bind(this))
        .style(this.style)
        .on("mouseover", this.mouseOverBox.bind(this))
        .on("mouseout", this.mouseOutBox.bind(this));

    bars.selectAll("text").remove();
    bars.append("text");

    var text = bars.selectAll("text");
    text
        .attr("x", function(d) { return this.margin.left + this.xScale(d.x) }.bind(this))
        .attr("y", function(d) { return this.margin.top + this.yScale(d.y)}.bind(this))
        .attr("dy", "20px")
        .attr("dx", function(d) { return this.xScale(this.xScale.domain()[0] + (d.dx/2)); }.bind(this))
        .attr("fill", "#fff")
        .attr("text-anchor", "middle")
        .attr("pointer-events", "none")
        .text(function(d) { return this.tickFormat(d.y) }.bind(this));

}

/** 
 * Adds the specified data set to this graph
 * @param data The data to be added to this graph
 * @param style A set of css attributes that will be applied to each rectangle (bar) on the chart
 * @param xAccessor A function which will return the x value of a point when called with the point as a parameter
 * @param yAccessor A function which will return the y value of a point when called with the point as a parameter
 */
safe_graph_barchart.prototype.setData = function(data, style, xAccessor, yAccessor){
    if(data == undefined){ return }

    this.data = data;
    this.xAccessor = xAccessor;
    this.yAccessor = yAccessor;
    this.style = style;

    this.display();
}

safe_graph_barchart.prototype.update = function(){
    this.display();
}

safe_graph_barchart.prototype.getData = function(){
    return this.data;
}

safe_graph_barchart.prototype.getExtent = function(){
    return this.extent;
}

safe_graph_barchart.prototype.remove = function(){
    this.svg.selectAll(".bar").remove();
}

safe_graph_barchart.prototype.setBinNumber = function(bins){
    this.bins = bins;
    this.update();
}

safe_graph_barchart.prototype.getBinNumber = function(){
    return this.bins;
}

safe_graph_barchart.prototype.setBinWidth = function(width){
    this.bins = Math.round(d3.max(this.data) / width);
    if(this.bins <= 0){ this.bins = 1 }
    this.update();
}

safe_graph_barchart.prototype.getBinWidth = function(){
    if(!(this.histogram && this.histogram.length > 0)){ return 0 }
    return this.histogram[0].dx;
}

safe_graph_barchart.prototype.setFrequency = function(frequency){
    this.frequency = frequency;
    this.update();
}

safe_graph_barchart.prototype.getFrequency = function(){
    return this.frequency;
}

/** Called when mouse is over a histogram box */
safe_graph_barchart.prototype.mouseOverBox = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0.9);

    this.tooltip.html(d3.event.srcElement.getAttribute("info"))
        .style("left", (d3.event.pageX) + "px")
        .style("top", (d3.event.pageY - 28) + "px");
}

/** Called when mouse is no longer over a histogram box */
safe_graph_barchart.prototype.mouseOutBox = function(){
    this.tooltip.transition()
        .duration(200)
        .style("opacity", 0);
}
