/**-------------------------------------
 * A selector for the lineplot graphs
 * -------------------------------------
 * 
 * A selector which allows the selection of one level for each
 * factor except one. That factor is the one to be varied as
 * others remain constant
 */
safe.selector.lineplot = safe_selector_lineplot;

/**
 * Create a new selector object
 *
 * @param exp_id The id of the experiment currently being observed
 * @param factor The factor that is to be varied as others remain constant
 * @param metricSelectCallBack The function to be called when a metric is selected
 * @param parentElement The dom element, d3 object, or dom id of the element
 *          to which this element should be appended
 */
function safe_selector_lineplot(exp_id, factor, metricSelectCallBack, parentElement){
	this.exp_id = exp_id;
	this.factor = factor;
	this.metricSelectCallBack = metricSelectCallBack;
    this.dParent = safe.tools.d3Object(parentElement);

	this.dSelf = this.dParent.append("div")
        .attr("class", "container well")
        .style("width", "95%");

    var hider = new safe.tools.hider(this.dSelf, this.dParent);
    hider.initComponents();

    this.levelList = {};
    this.factorList = {};
    this.parameters = {};
    this.selectMetric = new safe.selector.metric(this.exp_id, this.metricSelectChange.bind(this), this.dSelf);

    this.selectMetric.readInitialData(function(){
        this.readInitialData(this.initComponents.bind(this));
    }.bind(this));
}

/**
 * Reads the initial data required for proper operation
 * @param callBack A function to be called after the read is complete
 */
safe_selector_lineplot.prototype.readInitialData = function(callBack){
    d3.json("/safe/exp/" + this.exp_id + "/levels/", function(error, data){
        if(error) { console.warn(error) }

        Object.keys(data).forEach(function(factor_id){
            if(factor_id == this.factor.id){ return }
            this.factorList[factor_id] = data[factor_id];
            this.levelList[factor_id] = data[factor_id].levels;
            this.parameters[data[factor_id].name] = data[factor_id].levels[0];
        }.bind(this));
        if(callBack != undefined){ callBack() }
    }.bind(this));
}

/**
 * Initializes the dom elmenets of this object
 */
safe_selector_lineplot.prototype.initComponents = function(){
    var selectors = this.dSelf.selectAll(".factor").data(Object.keys(this.levelList)).enter()
        .append("div")
        .attr("class", "col-xs-12 factor");

    selectors.append("label")
        .attr("class", "col-xs-6")
        .text(function(d) { return this.factorList[d].name + " : " }.bind(this));

    selectors
        .append("select")
        .attr("class", "form-control")
        .attr("factor_id", function(d){ return d })
        .on("change", this.levelSelectChange.bind(this))
            .selectAll("option").data(function(d){ return this.levelList[d] }.bind(this))
            .enter()
            .append("option")
            .text(function(d){ return d });

    this.dSelf.append("div").attr("class", "clearfix");

    this.selectMetric.initComponents();
    this.selectMetric.setDistinctionId(getKeyString(this.parameters));
}

/**
 * Called when a selected level is changed
 */
safe_selector_lineplot.prototype.levelSelectChange = function(){
    var select = d3.event.target;
    var level = select.options[select.selectedIndex].value;
    var factor_name = this.factorList[select.getAttribute("factor_id")].name;
    this.parameters[factor_name] = level;
    this.selectMetric.setDistinctionId(getKeyString(this.parameters));
}

/**
 * Called when a metric is selected
 */
safe_selector_lineplot.prototype.metricSelectChange = function(metric){
    this.metricSelectCallBack(this.parameters, metric);
}

/**
 * Set a metric as being selected 
 * @param keyString A string that destinguishes one set of metrics from another
 *          An example could be a run id
 * @param metric_id The id of the metric that is to be set as selected
 * @param color The color to be used in indicating this metric as selected
 */
safe_selector_lineplot.prototype.addDisplayed = function(keyString, metric_id, color){
    this.selectMetric.setMetricColor(keyString, metric_id, color);
    this.selectMetric.updateSelectedMetrics();
}

/**
 * Set a metric as being unselected
 * @param keyString A string that destinguishes one set of metrics from another
 *          An example could be a run id
 * @param metric_id The id of the metric that is to be set as unselected
 */
safe_selector_lineplot.prototype.removeDisplayed = function(keyString, metric_id){
    this.selectMetric.setMetricColor(keyString, metric_id, undefined);
    this.selectMetric.updateSelectedMetrics();
}

