/**-------------------------------------
 * A designpoint selector object
 * -------------------------------------
 *
 * An object prototype which allows the selection of design points
 * which are stored in the database.
 */

safe.selector.designpoint = safe_selector_designpoint;

/**
 * Creates a new selection object
 * @param exp_id The experiment id with which this selector is associated
 * @param changeCallBack A function to be called when the selection is changed
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_designpoint(exp_id, changeCallBack, parentElement){
    this.exp_id = exp_id;
    this.callBack = changeCallBack;
    this.dParent = safe.tools.d3Object(parentElement);

    this.designPointList = {};
    this.designPointFactors = undefined;
    this.parameters = {};
}

/**
 * Reads data from the database about available design points
 * @param callBack A function to be called when the read is complete
 */
safe_selector_designpoint.prototype.readInitialData = function(callBack){
    this.designPointFactors = {};
    d3.json("/safe/exp/" + this.exp_id + "/designpoints/", function(error, designPoints){
        designPoints.forEach(function(designPoint){
            this.designPointFactors[designPoint.id] = designPoint.factors;
        }.bind(this));
        if(callBack != undefined){ callBack() }
    }.bind(this));
}

/**
 * Reads data from the database about design points associated
 * with the selection parameters currently in place for the selector
 *
 * @param callBack A functio to be called when the read is complete
 */
safe_selector_designpoint.prototype.readData = function(callBack){
    var update = function(){ 
        var parameters = this.getParameters();

        d3.json("/safe/exp/" + this.exp_id + "/dp-range-query/").post(JSON.stringify(parameters), function(error, designPoints){
            this.designPointList = designPoints;
            if(callBack != undefined){ callBack() };
        }.bind(this));
    }.bind(this)
    
    if(this.designPointFactors == undefined ){
        this.readInitialData(update.bind(this));
    }else{
      update();
    }
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_designpoint.prototype.initComponents = function(){
    this.dSelf = this.dParent
        .append("div");

    this.dSelect = this.dSelf
        .append("select")
        .attr("class", "form-control")
        .on("change", this.selectChange.bind(this));
    this.eSelect = this.dSelect[0][0];

    this.dInfo = this.dSelf
        .append("div");
}

/**
 * Updates the currently displayed design points based on data in the database
 * @param callBack A function to be called after the update is complete
 */
safe_selector_designpoint.prototype.updateDesignPoints = function(callBack){
    var update = function(){
        designPoints = this.dSelect.selectAll("option").data(this.designPointList);
        designPoints.enter().append("option");
        designPoints.attr("designPoint", function(d) { return d })
            .text(function(d) { return "DesignPoint " + d });
        designPoints.exit().remove();

        this.updateDesignPointInfo();
        
        if(callBack != undefined){ callBack() }
    }.bind(this);
    this.readData(update.bind(this));
}

/**
 * Updates the display which holds information about the currently 
 * selected design point.
 */
safe_selector_designpoint.prototype.updateDesignPointInfo = function(){
    var designPoint = this.designPointFactors[this.getSelectedDesignPoint()];

    var lines = this.dInfo.selectAll("label").data(Object.keys(designPoint))
    lines.enter().append("label");
    lines.attr("class", "col-xs-12").text(function(d) { return d + " : " + designPoint[d] });
    lines.exit().remove();
    lines.sort();
}

/**
 * Called when there is a change in the currently selected design point.
 */
safe_selector_designpoint.prototype.selectChange = function(){
    this.updateDesignPointInfo();
    this.callBack(this.getSelectedDesignPoint());
}

/**
 * Returns the design point that is currently selected
 * @return The currently selected design point
 */
safe_selector_designpoint.prototype.getSelectedDesignPoint = function(){
    designpoint_id = this.eSelect.options[this.eSelect.selectedIndex].getAttribute("designpoint");
    return designpoint_id;
}

/**
 * Adds a parameter for the reading of design point data
 * @param key A distinct string with which to associate the parameter
 * @param parameter A restriction of the form {"factor_name": <factor_name>, "min_value":<min_value>, "max_value":<max_value>}
 */
safe_selector_designpoint.prototype.addParameter = function(key, parameter){
    this.parameters[key] = parameter;
}

/**
 * Removes a parameter that was previously added
 * @param key The key associated with the parameter through "addParameter"
 */ 
safe_selector_designpoint.prototype.removeParameter = function(key){
    delete this.parameters[key];
}

/**
 * Returns a list of all parameters being used by this selector
 * @return A list of all parameters of the form
 *      [{"factor_name": <factor_name>, "min_value":<min_value>, "max_value":<max_value>}, ...]
 */
safe_selector_designpoint.prototype.getParameters = function(){
    parameters = [];
    Object.keys(this.parameters).forEach(function(parameterKey){
        parameters.push(this.parameters[parameterKey]);
    }.bind(this));
    return parameters;
}
