/**-------------------------------------
 * A graph options object
 * -------------------------------------
 *
 * An object prototype which allows the selection of various
 * setting for a currently selected graph
 */
safe.selector.options = safe_selector_options;

/**
 * Creates a new selection object
 * @param exp_id The experiment id with which this selector is associated
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_options(titleElement, parentElement){
    this.titleElement = safe.tools.d3Object(titleElement);
    this.dParent = safe.tools.d3Object(parentElement);

    this.dSelf = this.dParent.append("div")
        .attr("class", "container well")
        .style("width", "95%");
    this.eSelf = this.dSelf[0][0];

    this.initComponents();
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_options.prototype.initComponents = function(){
    var hider = new safe.tools.hider(this.dSelf, this.dParent);
    hider.initComponents();

    this.dSelf.append("label")
        .attr("class", "col-xs-12") 
        .text("Chart Title:");

    var chartTitle = this.dSelf.append("input")
        .attr("class", "form-control")
        .attr("value", this.titleElement[0][0].innerText)
        .on("change", this.titleChange.bind(this));

    var xAxis = this.dSelf.append("div");
    xAxis.append("label").attr("class", "col-xs-12").text("X Axis Label:");
    var xAxisLabel = xAxis.append("input").attr("class", "form-control").on("change", this.labelChange.bind(this));

    var yAxis = this.dSelf.append("div");
    yAxis.append("label").attr("class", "col-xs-12").text("Y Axis Label:");
    var yAxisLabel = yAxis.append("input").attr("class", "form-control").on("change", this.labelChange.bind(this));

    this.dSelf.append("label").attr("class", "col-xs-12").text("Selected Graph Height:");
    var selectedHeight = this.dSelf.append("input")
        .attr("class", "form-control")
        .attr("type", "number")
        .attr("min", 0)
        .on("change", this.selectedHeightChange.bind(this));

    this.dSelf.append("label").attr("class", "col-xs-12").text("Context Graph Height:");
    var contextHeight = this.dSelf.append("input")
        .attr("class", "form-control")
        .attr("type", "number")
        .attr("min", 0)
        .on("change", this.contextHeightChange.bind(this));

    this.dSelf.append("label").attr("class", "col-xs-12").text("Move Selected Graph:");
    var moveUp = this.dSelf.append("button").attr("class", "btn btn-default col-xs-6").text("UP").on("click", this.moveSelectedUp.bind(this));
    var moveDown = this.dSelf.append("button").attr("class", "btn btn-default col-xs-6").text("DOWN").on("click", this.moveSelectedDown.bind(this));

    this.dSelf.append("label").attr("class", "col-xs-12").text("Download Data:");
    var downloadData = this.dSelf.append("button")
        .attr("class", "btn btn-default col-xs-12")
        .text("Download JSON Data")
        .on("click", this.downloadData.bind(this));

    var downloadSvg = this.dSelf.append("button")
        .attr("class", "btn btn-default col-xs-12")
        .text("Download Graph SVG")
        .on("click", function(){ if(this.selectedGraph){ safe.tools.downloadElement(this.selectedGraph.svg[0][0], "graph.svg")} }.bind(this));

    this.chartTitle = chartTitle[0][0];
    this.xAxisLabel = xAxisLabel[0][0];
    this.yAxisLabel = yAxisLabel[0][0];
    this.selectedGraphHeight = selectedHeight[0][0];
    this.contextGraphHeight = contextHeight[0][0];
    this.moveUp = moveUp[0][0];
    this.moveDown = moveDown[0][0];
    this.downloadData = downloadData[0][0];
    this.downloadSvg = downloadSvg[0][0];

    this.contextGraphHeight.disabled = true;
    this.selectedGraphHeight.disabled = true;
}

/**
 * Sets an object to be treated as the curretnly selected graph
 * @param graph The graph to be treated as the currently selected graph
 */
safe_selector_options.prototype.setSelectedGraph = function(graph){
    this.selectedGraph = graph;
    this.selectedGraphHeight.value = graph.getHeight();
    this.setLabels();
    this.selectedGraphHeight.disabled = false;
}

/**
 * Sets an object to be treated as the context graph
 * @param graph The graph to be treated as the context graph
 */
safe_selector_options.prototype.setContextGraph = function(graph){
    this.contextGraph = graph;
    this.contextGraphHeight.value = graph.getHeight();
    this.contextGraphHeight.disabled = false;
}

/**
 * Sets the title of the grpah to the currently set title
 */
safe_selector_options.prototype.titleChange = function(){
    if(!this.titleElement){ return }
    this.titleElement.text(this.chartTitle.value);
}


/**
 * Called when a label text box has been changed
 */
safe_selector_options.prototype.labelChange = function(){
    if(this.selectedGraph){
        this.selectedGraph.setXAxisLabel(this.xAxisLabel.value);
        this.selectedGraph.setYAxisLabel(this.yAxisLabel.value);
    }
    if(this.contextGraph){
        this.contextGraph.setXAxisLabel(this.xAxisLabel.value);
        this.contextGraph.setYAxisLabel(this.yAxisLabel.value);
    }
}

/**
 * Sets the label text boxes based on the currently selected graph
 */
safe_selector_options.prototype.setLabels = function(){
    if(this.selectedGraph){
        this.xAxisLabel.value = this.selectedGraph.getXAxisLabel();
        this.yAxisLabel.value = this.selectedGraph.getYAxisLabel();
    }
    if(this.contextGraph){
        this.contextGraph.setXAxisLabel(this.selectedGraph.getXAxisLabel());
        this.contextGraph.setYAxisLabel(this.selectedGraph.getYAxisLabel());
    }
}

/**
 * Switches the positions of the currently selected graph and its
 * previous sibling in the dom tree. Also moves the context to its
 * correct position.
 */
safe_selector_options.prototype.moveSelectedUp = function(){
    if(!this.selectedGraph){ return }

    var selected = this.selectedGraph.svg[0][0];
    selected.parentNode.insertBefore(selected, selected.previousSibling);

    if(this.contextGraph){
        var context = this.contextGraph.svg[0][0];
        context.parentNode.insertBefore(context, context.previousSibling);
    }
}

/**
 * Switches the positions of the currently selected graph and its
 * next sibling in the dom tree. Also moves the context to its
 * correct position.
 */
safe_selector_options.prototype.moveSelectedDown = function(){
    if(!(this.selectedGraph)){ return }

    var selected = this.selectedGraph.svg[0][0];

    if(this.contextGraph){
        var context = this.contextGraph.svg[0][0];
        if(context.nextSibling){ 
            context.parentNode.insertBefore(context, context.nextSibling.nextSibling);
            selected.parentNode.insertBefore(selected, selected.nextSibling.nextSibling);
        }else{
            context.parentNode.insertBefore(context, context.parentNode.firstChild);
            selected.parentNode.insertBefore(selected, selected.parentNode.firstChild);
        }
   }else{
        if(selected.nextSibling){
            selected.parentNode.insertBefore(selected, selected.nextSibling.nextSibling);
        }else{
            selected.parentNode.insertBefore(selected, selected.parentNode.firstChild);
        }
   }
}

/**
 * Called when the selected graph height selector is changed and reasigns the
 * height of the graph accordingly.
 */
safe_selector_options.prototype.selectedHeightChange = function(){
    if(this.selectedGraph){
        this.selectedGraph.setHeight(parseInt(this.selectedGraphHeight.value));
        this.selectedGraph.displayAll();
    }
}

/**
 * Called when the context graph height selector is changed and reasigns the
 * height of the graph accordingly.
 */
safe_selector_options.prototype.contextHeightChange = function(){
    if(this.contextGraph){
        this.contextGraph.setHeight(parseInt(this.contextGraphHeight.value));
        this.contextGraph.displayAll();

        this.contextGraph.svg.selectAll(".brush").selectAll("rect")
            .attr("y", -6)
            .attr("height", parseInt(this.contextGraphHeight.value - this.contextGraph.margin.bottom - this.contextGraph.margin.top) + 7);
    }
}

/**
 * Makes that the user can no longer move graphs up and down
 */
safe_selector_options.prototype.disableGraphMovement = function(){
    this.moveUp.disabled = true;
    this.moveDown.disabled = true;
}

/**
 * Ensures that the user can move graphs up and down
 */
safe_selector_options.prototype.enableGraphMovement = function(){
    this.moveUp.disabled = false;
    this.moveDown.disabled = false;
}

safe_selector_options.prototype.downloadData = function(){
    var data = "text/json;charset=utf-8," + encodeURIComponent(JSON.stringify(this.selectedGraph.getData()));
    //var link = d3.select("#container").append("a")
    //    .attr("href", "data:" + data)
    //    .attr("download", "data.json");
    //var element = link[0][0];
    var element = document.createElement("a");
    element.setAttribute("href", "data:" + data);
    element.setAttribute("download", "data.json");
    element.click();
    //$('<a id="dataDownload" href="data:' + data + '" download="data.json"></a>').appendTo('#container');
    //$("#dataDownload")[0].click();
}
